from udsoncan import Request, Response
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseSubfunction, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
import udsoncan.tools as tools

from typing import Optional, cast


class ECUReset(BaseService):
    _sid = 0x11

    supported_negative_response = [ResponseCode.SubFunctionNotSupported,
                                   ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.SecurityAccessDenied
                                   ]

    class ResetType(BaseSubfunction):
        """
        ECUReset defined subfunctions
        """
        __pretty_name__ = 'reset type'  # Only to print "custom reset type" instead of "custom subfunction"

        hardReset = 1
        keyOffOnReset = 2
        softReset = 3
        enableRapidPowerShutDown = 4
        disableRapidPowerShutDown = 5

    class ResponseData(BaseResponseData):
        """
        .. data:: reset_type_echo

                Request subfunction echoed back by the server

        .. data:: powerdown_time

                Amount of time, in seconds, before the power down sequence is executed. Should be provided only when reset type is enableRapidPowerShutDown
        """

        reset_type_echo: int
        powerdown_time: Optional[int]

        def __init__(self, reset_type_echo: int, powerdown_time: Optional[int] = None):
            super().__init__(ECUReset)

            self.reset_type_echo = reset_type_echo
            self.powerdown_time = powerdown_time

    class InterpretedResponse(Response):
        service_data: "ECUReset.ResponseData"

    @classmethod
    def make_request(cls, reset_type: int) -> Request:
        """
        Generates a request for ECUReset

        :param reset_type: Service subfunction. Allowed values are from 0 to 0x7F
        :type reset_type: int

        :raises ValueError: If parameters are out of range, missing or wrong type
        """
        tools.validate_int(reset_type, min=0, max=0x7F, name='Reset type')
        return Request(service=cls, subfunction=reset_type)

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`ECUReset.ResponseData<udsoncan.services.ECUReset.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :raises InvalidResponseException: If length of ``response.data`` is too short
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 1: 	# Should not happen as response decoder will raise an exception.
            raise InvalidResponseException(response, "Response data must be at least 1 bytes")

        response.service_data = cls.ResponseData(
            reset_type_echo=response.data[0]
        )

        if response.service_data.reset_type_echo == cls.ResetType.enableRapidPowerShutDown:
            if len(response.data) < 2:
                raise InvalidResponseException(response, 'Response data is missing a second byte for reset type "enableRapidPowerShutDown"')

            response.service_data.powerdown_time = response.data[1]

        return cast(ECUReset.InterpretedResponse, response)



from udsoncan.client import Client
from udsoncan import services
from udsoncan.exceptions import *
from udsoncan import DidCodec, AsciiCodec
import struct

from test.ClientServerTest import ClientServerTest


class StubbedDidCodec(DidCodec):
    def encode(self, did_value):
        return struct.pack('B', did_value + 1)

    def decode(self, did_payload):
        return struct.unpack('B', did_payload)[0] - 1

    def __len__(self):
        return 1


class ReadRemainingDataCodec(DidCodec):

    def encode(self, *args, **kwargs):
        return b''

    def decode(self, did_payload):
        return did_payload

    def __len__(self):
        raise self.ReadAllRemainingData


class CodecWithNoLength(DidCodec):

    def encode(self, *args, **kwargs):
        return b''

    def decode(self, did_payload):
        return did_payload


class TestReadDataByIdentifier(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def postClientSetUp(self):
        self.udsclient.config["data_identifiers"] = {
            1: '>H',
            2: '<H',
            3: StubbedDidCodec,
            4: AsciiCodec(5),
            5: ReadRemainingDataCodec,
            6: CodecWithNoLength
        }

    def test_rdbi_single_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x22\x00\x01")
        self.conn.fromuserqueue.put(b"\x62\x00\x01\x12\x34")  # Positive response

    def _test_rdbi_single_success(self):
        response = self.udsclient.read_data_by_identifier(didlist=1)
        self.assertTrue(response.positive)
        values = response.service_data.values
        self.assertEqual(values[1], (0x1234,))

    def test_rdbi_single_success_spr_no_effect(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x22\x00\x01")
        self.conn.fromuserqueue.put(b"\x62\x00\x01\x12\x34")  # Positive response

    def _test_rdbi_single_success_spr_no_effect(self):
        with self.udsclient.suppress_positive_response:
            response = self.udsclient.read_data_by_identifier(didlist=1)
            self.assertTrue(response.positive)
            values = response.service_data.values
            self.assertEqual(values[1], (0x1234,))

    def test_rdbi_multiple_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x22\x00\x01\x00\x02\x00\x04\x00\x03")
        self.conn.fromuserqueue.put(b"\x62\x00\x01\x12\x34\x00\x02\x56\x78\x00\x04\x61\x62\x63\x64\x65\x00\x03\x11")  # Positive response

    def _test_rdbi_multiple_success(self):
        response = self.udsclient.read_data_by_identifier(didlist=[1, 2, 4, 3])
        self.assertTrue(response.positive)
        values = response.service_data.values
        self.assertEqual(values[1], (0x1234,))
        self.assertEqual(values[2], (0x7856,))
        self.assertEqual(values[3], 0x10)
        self.assertEqual(values[4], 'abcde')

    def test_rdbi_multiple_zero_padding1_success(self):
        data = b'\x62\x00\x01\x12\x34\x00\x02\x56\x78\x00\x03\x11'
        for i in range(8):
            self.wait_request_and_respond(data + b"\x00" * (i + 1))

    def _test_rdbi_multiple_zero_padding1_success(self):
        self.udsclient.config['tolerate_zero_padding'] = True
        for i in range(8):
            response = self.udsclient.read_data_by_identifier(didlist=[1, 2, 3])
            self.assertTrue(response.positive)
            values = response.service_data.values
            self.assertEqual(values[1], (0x1234,))
            self.assertEqual(values[2], (0x7856,))
            self.assertEqual(values[3], 0x10)
            self.assertFalse(0 in values)

    def test_rdbi_multiple_zero_padding_not_tolerated_exception(self):
        data = b'\x62\x00\x01\x12\x34\x00\x02\x56\x78\x00\x03\x11'

        self.wait_request_and_respond(data + b"\x00")  # One extra byte is incomplete DID = invalid response

        for i in range(1, 7):
            # 2 extra bytes = valid DID. Make sure to have data in payload as library do not allow empty pack string
            self.wait_request_and_respond(data + b"\x00" * (i + 2))

    def _test_rdbi_multiple_zero_padding_not_tolerated_exception(self):
        self.udsclient.config['tolerate_zero_padding'] = False
        for i in range(1):
            with self.assertRaises(InvalidResponseException):
                self.udsclient.read_data_by_identifier(didlist=[1, 2, 3])

        for i in range(1, 7):
            with self.assertRaises(UnexpectedResponseException):  # Not requested DID 0x0000
                self.udsclient.config['data_identifiers'][0] = 'B' * i
                self.udsclient.read_data_by_identifier(didlist=[1, 2, 3])

    def test_rdbi_multiple_zero_padding_not_tolerated_no_exception(self):
        data = b'\x62\x00\x01\x12\x34\x00\x02\x56\x78\x00\x03\x11'
        self.wait_request_and_respond(data + b"\x00")  # One extra byte is incomplete DID = invalid response

        for i in range(1, 7):
            # 2 extra bytes = valid DID. Make sure to have data in payload as library do not allow empty pack string
            self.wait_request_and_respond(data + b"\x00" * (i + 2))

    def _test_rdbi_multiple_zero_padding_not_tolerated_no_exception(self):
        self.udsclient.config['tolerate_zero_padding'] = False

        self.udsclient.config['exception_on_invalid_response'] = False
        self.udsclient.config['exception_on_unexpected_response'] = True

        response = self.udsclient.read_data_by_identifier(didlist=[1, 2, 3])
        self.assertFalse(response.valid)

        self.udsclient.config['exception_on_invalid_response'] = True
        self.udsclient.config['exception_on_unexpected_response'] = False
        for i in range(1, 7):
            self.udsclient.config['data_identifiers'][0] = 'B' * i
            response = self.udsclient.read_data_by_identifier(didlist=[1, 2, 3])
            self.assertTrue(response.valid)
            self.assertTrue(response.unexpected)

    def test_rdbi_variable_size_did(self):
        self.wait_request_and_respond(b"\x62\x00\x01\x12\x34\x00\x02\x56\x78\x00\x05\xaa\xbb\xcc\xdd")

    def _test_rdbi_variable_size_did(self):
        response = self.udsclient.read_data_by_identifier(didlist=[1, 2, 5])
        self.assertTrue(response.positive)
        values = response.service_data.values
        self.assertEqual(values[1], (0x1234,))
        self.assertEqual(values[2], (0x7856,))
        self.assertEqual(values[5], b'\xaa\xbb\xcc\xdd')

    # DID 5 read all the data up to the end. Makes no sense to read another DID after that.
    def test_rdbi_variable_size_did_not_last(self):
        pass

    def _test_rdbi_variable_size_did_not_last(self):
        with self.assertRaises(ValueError):
            self.udsclient.read_data_by_identifier(didlist=[1, 2, 5, 3])

    def test_rdbi_incomplete_response_exception(self):
        self.wait_request_and_respond(b"\x62\x00\x01\x12\x34\x00\x02\x56\x78\x00\x03")

    def _test_rdbi_incomplete_response_exception(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.read_data_by_identifier(didlist=[1, 2, 3])

    def test_rdbi_incomplete_response_no_exception(self):
        self.wait_request_and_respond(b"\x62\x00\x01\x12\x34\x00\x02\x56\x78\x00\x03")

    def _test_rdbi_incomplete_response_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.read_data_by_identifier(didlist=[1, 2, 3])
        self.assertFalse(response.valid)

    def test_rdbi_unknown_did_exception(self):
        self.wait_request_and_respond(b"\x62\x00\x09\x12\x34\x00\x02\x56\x78\x00\x03\x11")

    def _test_rdbi_unknown_did_exception(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.read_data_by_identifier(didlist=[1, 2, 3])

    def test_rdbi_unknown_did_no_exception(self):
        self.wait_request_and_respond(b"\x62\x00\x09\x12\x34\x00\x02\x56\x78\x00\x03\x11")

    def _test_rdbi_unknown_did_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.read_data_by_identifier(didlist=[1, 2, 3])
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_rdbi_unwanted_did_exception(self):
        self.wait_request_and_respond(b"\x62\x00\x01\x12\x34\x00\x02\x56\x78\x00\x03\x11")

    def _test_rdbi_unwanted_did_exception(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.read_data_by_identifier(didlist=[1, 3])

    def test_rdbi_unwanted_did_no_exception(self):
        self.wait_request_and_respond(b"\x62\x00\x01\x12\x34\x00\x02\x56\x78\x00\x03\x11")

    def _test_rdbi_unwanted_did_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.read_data_by_identifier(didlist=[1, 3])
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_rdbi_invalidservice_exception(self):
        self.wait_request_and_respond(b"\x00\x00\x01\x12\x34")  # Service is inexistant

    def _test_rdbi_invalidservice_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.read_data_by_identifier(didlist=1)

    def test_rdbi_invalidservice_no_exception(self):
        self.wait_request_and_respond(b"\x00\x00\x01\x12\x34")  # Service is inexistant

    def _test_rdbi_invalidservice_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.read_data_by_identifier(didlist=1)
        self.assertFalse(response.valid)

    def test_rdbi_wrongservice_exception(self):
        self.wait_request_and_respond(b"\x50\x00\x01\x12\x34")  # Valid service, but not the one requested

    def _test_rdbi_wrongservice_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.read_data_by_identifier(didlist=1)

    def test_rdbi_wrongservice_no_exception(self):
        self.wait_request_and_respond(b"\x50\x00\x01\x12\x34")  # Valid service, but not the one requested

    def _test_rdbi_wrongservice_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.read_data_by_identifier(didlist=1)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_no_config(self):
        pass

    def _test_no_config(self):
        with self.assertRaises(ConfigError):
            self.udsclient.read_data_by_identifier(didlist=[1, 2, 3, 99])

    def test_no_length(self):
        pass

    def _test_no_length(self):
        with self.assertRaises(NotImplementedError):
            self.udsclient.read_data_by_identifier(didlist=[6])



from udsoncan.exceptions import ConfigError
from udsoncan.BaseService import BaseService
from copy import copy

from udsoncan.typing import IOConfigEntry

from typing import Any, Union, List, Dict, Type, Optional


def validate_int(value: Any, min: int = 0, max: int = 0xFF, name: str = 'value'):
    if not isinstance(value, int):
        raise ValueError("%s must be a valid integer" % (name))
    if value < min or value > max:
        raise ValueError("%s must be an integer between 0x%X and 0x%X" % (name, min, max))


# Make sure that the actual client configuration contains valid definitions for given Input/Output Data Identifiers
def check_io_config(didlist: Union[int, List[int]], ioconfig: Dict[Any, Any]) -> Dict[int, IOConfigEntry]:
    didlist = [didlist] if not isinstance(didlist, list) else didlist
    ioconfig = copy(ioconfig)
    if 'input_output' in ioconfig:
        ioconfig = ioconfig['input_output']

    if not isinstance(ioconfig, dict):
        raise ConfigError('input_output', msg='Configuration of Input/Output section must be a dict.')

    for did in didlist:
        if did not in ioconfig:
            raise ConfigError(key=did, msg='Actual Input/Output configuration contains no definition for data identifier 0x%04x' % did)
        if isinstance(ioconfig[did], dict):  # IO Control services has that concept of composite DID. We define them with dicts.
            if 'codec'not in ioconfig[did]:
                raise ConfigError('codec', msg='Configuration for Input/Output identifier 0x%04x is missing a codec')

            if 'mask' in ioconfig[did]:
                mask_def = ioconfig[did]['mask']
                for mask_name in mask_def:
                    if not isinstance(mask_def[mask_name], int):
                        raise ValueError('In Input/Output configuration for did 0x%04x, mask "%s" is not an integer' % (did, mask_name))

                    if mask_def[mask_name] < 0:
                        raise ValueError('In Input/Output configuration for did 0x%04x, mask "%s" is not a positive integer' % (did, mask_name))

            if 'mask_size' in ioconfig[did]:
                if not isinstance(ioconfig[did]['mask_size'], int):
                    raise ValueError('mask_size in Input/Output configuration for did 0x%04x must be a valid integer' % (did))

                if ioconfig[did]['mask_size'] < 0:
                    raise ValueError('mask_size in Input/Output configuration for did 0x%04x must be greater than 0' % (did))

                if 'mask' in ioconfig[did]:
                    mask_def = ioconfig[did]['mask']
                    for mask_name in mask_def:
                        if mask_def[mask_name] > 2**(ioconfig[did]['mask_size'] * 8) - 1:
                            raise ValueError(
                                'In Input/Output configuration for did 0x%04x, mask "%s" cannot fit in %d bytes (defined by mask_size)' % (did, mask_name, ioconfig[did]['mask_size']))

        else:
            ioconfig[did] = {
                'codec': ioconfig[did]
            }

    return ioconfig


def cls_from_request_id(given_id: int) -> Optional[Type[BaseService]]:
    return BaseService.from_request_id(given_id)


def is_valid_service(service_cls: Type) -> bool:
    return issubclass(service_cls, BaseService)



from udsoncan.client import Client
from udsoncan import services, DataFormatIdentifier, Filesize
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest

class TestRequestFileTransfer(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    #============== Add File ================

    def test_add_file_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x38\x01\x00\x0b" + "my_file.txt".encode('ascii') + b"\x52\x04\x00\x00\x02\x22\x00\x00\x01\x11")
        self.conn.fromuserqueue.put(b"\x78\x01\x02\xab\xcd\x52")	# Positive response

    def _test_add_file_success(self):
        response = self.udsclient.add_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2), filesize = Filesize(uncompressed=0x222, compressed=0x111, width=4))
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.moop_echo, 1)
        self.assertEqual(response.service_data.max_length, 0xabcd)
        self.assertEqual(response.service_data.dfi.compression, 5)
        self.assertEqual(response.service_data.dfi.encryption, 2)
        self.assertIsNone(response.service_data.filesize)     # No filesize info when doing AddFile. Only for ReadFile
        self.assertIsNone(response.service_data.dirinfo_length)

    def test_add_file_bad_moop_echo(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x02\x02\xab\xcd\x00")

    def _test_add_file_bad_moop_echo(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.add_file("my_file.txt", filesize = 0x100)            

    def test_add_file_bad_moop_echo_no_exception(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x02\x02\xab\xcd\x00")

    def _test_add_file_bad_moop_echo_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False        
        response = self.udsclient.add_file("my_file.txt", filesize = 0x100)
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertTrue(response.unexpected)

    def test_add_file_bad_dfi(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x01\x02\xab\xcd\x11")

    def _test_add_file_bad_dfi(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.add_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2), filesize = 0x100)

    def test_add_file_zerolen_length(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x01\x00\x00")

    def _test_add_file_zerolen_length(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.add_file("my_file.txt", filesize = 0x100)

    def test_add_file_negative_response(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x7F\x38\x22")    # Condition not correct

    def _test_add_file_negative_response(self):
        with self.assertRaises(NegativeResponseException):
            self.udsclient.add_file("my_file.txt", filesize=0x100)

    def test_add_file_negative_response_no_exception(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x7F\x38\x22")    # Condition not correct

    def _test_add_file_negative_response_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.add_file("my_file.txt", filesize=0x100)
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)

    def test_add_file_invalid_length(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78")   
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x01")    
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x01\x02")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x01\x02\xab")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x01\x02\xab\xcd")

    def _test_add_file_invalid_length(self):
        for i in range(5):
            with self.assertRaises(InvalidResponseException):
                self.udsclient.add_file("my_file.txt", filesize = 0x100)

    def test_add_file_extra_bytes_response(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x01\x02\xab\xcd\x52\xAA")    # Positive response

    def _test_add_file_extra_bytes_response(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.add_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2), filesize = Filesize(uncompressed=0x222, compressed=0x111, width=4))

    def test_add_file_extra_bytes_response_zero_padding_ok(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x01\x02\xab\xcd\x52\x00")    # Positive response

    def _test_add_file_extra_bytes_response_zero_padding_ok(self):
        self.udsclient.config['tolerate_zero_padding'] = True
        response = self.udsclient.add_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2), filesize = Filesize(uncompressed=0x222, compressed=0x111, width=4))
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.moop_echo, 1)
        self.assertEqual(response.service_data.max_length, 0xabcd)
        self.assertEqual(response.service_data.dfi.compression, 5)
        self.assertEqual(response.service_data.dfi.encryption, 2)
        self.assertIsNone(response.service_data.filesize)     # No filesize info when doing AddFile. Only for ReadFile
        self.assertIsNone(response.service_data.dirinfo_length)

    def test_add_file_extra_bytes_response_zero_padding_not_ok(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x01\x02\xab\xcd\x52\x00")    # Positive response

    def _test_add_file_extra_bytes_response_zero_padding_not_ok(self):
        self.udsclient.config['tolerate_zero_padding'] = False
        with self.assertRaises(InvalidResponseException):
            self.udsclient.add_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2), filesize = Filesize(uncompressed=0x222, compressed=0x111, width=4))


    #============== Delete File ================

    def test_delete_file_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x38\x02\x00\x0b" + "my_file.txt".encode('ascii'))
        self.conn.fromuserqueue.put(b"\x78\x02")    # Positive response

    def _test_delete_file_success(self):
        response = self.udsclient.delete_file("my_file.txt")
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.moop_echo, 2)
        self.assertIsNone(response.service_data.max_length)
        self.assertIsNone(response.service_data.dfi)
        self.assertIsNone(response.service_data.filesize)
        self.assertIsNone(response.service_data.dirinfo_length)

    def test_delete_bad_moop_echo(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x03")    # Positive response

    def _test_delete_bad_moop_echo(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.delete_file("my_file.txt")

    def test_delete_bad_moop_echo_no_exception(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x03")    # Positive response

    def _test_delete_bad_moop_echo_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False        
        response = self.udsclient.delete_file("my_file.txt")
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertTrue(response.unexpected)

    def test_delete_file_negative_response(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x7F\x38\x22")    # Condition not correct

    def _test_delete_file_negative_response(self):
        with self.assertRaises(NegativeResponseException):
            self.udsclient.delete_file("my_file.txt")

    def test_delete_file_negative_response_no_exception(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x7F\x38\x22")    # Condition not correct

    def _test_delete_file_negative_response_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.delete_file("my_file.txt")
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)  

    def test_delete_file_extra_bytes_response(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x02\xAA")    # Positive response

    def _test_delete_file_extra_bytes_response(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.delete_file("my_file.txt")

    def test_delete_file_extra_bytes_response_zero_padding_ok(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x02\x00\x00")    # Positive response

    def _test_delete_file_extra_bytes_response_zero_padding_ok(self):
        self.udsclient.config['tolerate_zero_padding'] = True
        response = self.udsclient.delete_file("my_file.txt")
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.moop_echo, 2)
        self.assertIsNone(response.service_data.max_length)
        self.assertIsNone(response.service_data.dfi)
        self.assertIsNone(response.service_data.filesize)
        self.assertIsNone(response.service_data.dirinfo_length)

    def test_delete_file_extra_bytes_response_zero_padding_not_ok(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x02\x00\x00")    # Positive response

    def _test_delete_file_extra_bytes_response_zero_padding_not_ok(self):
        self.udsclient.config['tolerate_zero_padding'] = False
        with self.assertRaises(InvalidResponseException):
            self.udsclient.delete_file("my_file.txt")


    #============== Replace File ================

    def test_replace_file_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x38\x03\x00\x0b" + "my_file.txt".encode('ascii') + b"\x52\x04\x00\x00\x02\x22\x00\x00\x01\x11")
        self.conn.fromuserqueue.put(b"\x78\x03\x02\xab\xcd\x52")    # Positive response

    def _test_replace_file_success(self):
        response = self.udsclient.replace_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2), filesize = Filesize(uncompressed=0x222, compressed=0x111, width=4))
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.moop_echo, 3)
        self.assertEqual(response.service_data.max_length, 0xabcd)
        self.assertEqual(response.service_data.dfi.compression, 5)
        self.assertEqual(response.service_data.dfi.encryption, 2)
        self.assertIsNone(response.service_data.filesize)     # No filesize info when doing AddFile. Only for ReadFile
        self.assertIsNone(response.service_data.dirinfo_length)

    def test_replace_file_negative_response(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x7F\x38\x22")    # Condition not correct

    def _test_replace_file_negative_response(self):
        with self.assertRaises(NegativeResponseException):
            self.udsclient.replace_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2), filesize = Filesize(uncompressed=0x222, compressed=0x111, width=4))

    def test_replace_file_negative_response_no_exception(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x7F\x38\x22")    # Condition not correct

    def _test_replace_file_negative_response_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.replace_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2), filesize = Filesize(uncompressed=0x222, compressed=0x111, width=4))
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)           

    def test_replace_file_bad_moop_echo(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x02\x02\xab\xcd\x00")

    def _test_replace_file_bad_moop_echo(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.replace_file("my_file.txt", filesize = 0x100)            

    def test_replace_file_bad_moop_echo_no_exception(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x02\x02\xab\xcd\x00")

    def _test_replace_file_bad_moop_echo_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False        
        response = self.udsclient.replace_file("my_file.txt", filesize = 0x100)
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertTrue(response.unexpected)

    def test_replace_file_bad_dfi(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x03\x02\xab\xcd\x11")

    def _test_replace_file_bad_dfi(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.replace_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2), filesize = 0x100)

    def test_replace_file_zerolen_length(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x03\x00\x00")

    def _test_replace_file_zerolen_length(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.replace_file("my_file.txt", filesize = 0x100)

    def test_replace_file_zerolen_length_negative_response(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x7F\x38\x22")    # Condition not correct

    def _test_replace_file_zerolen_length_negative_response(self):
        with self.assertRaises(NegativeResponseException):
            self.udsclient.replace_file("my_file.txt", filesize=0x100)

    def test_replace_file_zerolen_length_negative_response_no_exception(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x7F\x38\x22")    # Condition not correct

    def _test_replace_file_zerolen_length_negative_response_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.replace_file("my_file.txt", filesize=0x100)
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)

    def test_replace_file_invalid_length(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78")   
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x03")    
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x03\x02")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x03\x02\xab")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x03\x02\xab\xcd")

    def _test_replace_file_invalid_length(self):
        for i in range(5):
            with self.assertRaises(InvalidResponseException):
                self.udsclient.replace_file("my_file.txt", filesize = 0x100)

    def test_replace_file_extra_bytes_response(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x03\x02\xab\xcd\x52\xAA")    # Positive response

    def _test_replace_file_extra_bytes_response(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.replace_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2), filesize = Filesize(uncompressed=0x222, compressed=0x111, width=4))              

    def test_replace_file_extra_bytes_response_zero_padding_ok(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x03\x02\xab\xcd\x52\x00\x00")    # Positive response

    def _test_replace_file_extra_bytes_response_zero_padding_ok(self):
        self.udsclient.config['tolerate_zero_padding'] = True
        response = self.udsclient.replace_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2), filesize = Filesize(uncompressed=0x222, compressed=0x111, width=4))              
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.moop_echo, 3)
        self.assertEqual(response.service_data.max_length, 0xabcd)
        self.assertEqual(response.service_data.dfi.compression, 5)
        self.assertEqual(response.service_data.dfi.encryption, 2)
        self.assertIsNone(response.service_data.filesize)     # No filesize info when doing AddFile. Only for ReadFile
        self.assertIsNone(response.service_data.dirinfo_length)

    def test_replace_file_extra_bytes_response_zero_padding_not_ok(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x03\x02\xab\xcd\x52\x00")    # Positive response

    def _test_replace_file_extra_bytes_response_zero_padding_not_ok(self):
        self.udsclient.config['tolerate_zero_padding'] = False
        with self.assertRaises(InvalidResponseException):
            self.udsclient.replace_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2), filesize = Filesize(uncompressed=0x222, compressed=0x111, width=4))              

    #============== Read File ================

    def test_read_file_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x38\x04\x00\x0b" + "my_file.txt".encode('ascii') + b"\x52")
        self.conn.fromuserqueue.put(b"\x78\x04\x02\xab\xcd\x52\x00\x02\x98\x76\x12\x34")    # Positive response

    def _test_read_file_success(self):
        response = self.udsclient.read_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2))
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.moop_echo, 4)
        self.assertEqual(response.service_data.max_length, 0xabcd)
        self.assertEqual(response.service_data.dfi.compression, 5)
        self.assertEqual(response.service_data.dfi.encryption, 2)
        self.assertEqual(response.service_data.filesize.uncompressed, 0x9876)
        self.assertEqual(response.service_data.filesize.compressed, 0x1234)
        self.assertIsNone(response.service_data.dirinfo_length)

    def test_read_file_negative_response(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x7F\x38\x22")    # Condition not correct

    def _test_read_file_negative_response(self):
        with self.assertRaises(NegativeResponseException):
            self.udsclient.read_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2))

    def test_read_file_negative_response_no_exception(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x7F\x38\x22")    # Condition not correct

    def _test_read_file_negative_response_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.read_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2))
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)           

    def test_read_file_bad_moop_echo(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x02\xab\xcd\x52\x00\x02\x98\x76\x12\x34")

    def _test_read_file_bad_moop_echo(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.read_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2))

    def test_read_file_bad_moop_echo_no_exception(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x02\xab\xcd\x52\x00\x02\x98\x76\x12\x34")

    def _test_read_file_bad_moop_echo_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False        
        response = self.udsclient.read_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2))
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertTrue(response.unexpected)

    def test_read_file_bad_dfi(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x02\xab\xcd\x11\x00\x02\x98\x76\x12\x34")

    def _test_read_file_bad_dfi(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.read_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2))

    def test_read_file_zerolen_length_lfi(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x00\x52\x00\x02\x98\x76\x12\x34")

    def _test_read_file_zerolen_length_lfi(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.read_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2))

    def test_read_file_zerolen_length_fsodipl(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x02\xab\xcd\x52\x00\x00")

    def _test_read_file_zerolen_length_fsodipl(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.read_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2))            

    def test_read_file_invalid_length(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x02")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x02\xab")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x02\xab\xcd")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x02\xab\xcd\x52")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x02\xab\xcd\x52\x00")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x02\xab\xcd\x52\x00\x02")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x02\xab\xcd\x52\x00\x02\x98")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x02\xab\xcd\x52\x00\x02\x98\x76")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x02\xab\xcd\x52\x00\x02\x98\x76\x12")

    def _test_read_file_invalid_length(self):
        for i in range(11):
            with self.assertRaises(InvalidResponseException):
                self.udsclient.read_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2))

    def test_read_file_extra_bytes_response(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x02\xab\xcd\x52\x00\x02\x98\x76\x12\x34\xAA")    # Positive response

    def _test_read_file_extra_bytes_response(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.read_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2))

    def test_read_file_extra_bytes_response_zero_padding_ok(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x02\xab\xcd\x52\x00\x02\x98\x76\x12\x34\x00")    # Positive response

    def _test_read_file_extra_bytes_response_zero_padding_ok(self):
        self.udsclient.config['tolerate_zero_padding'] = True
        response = self.udsclient.read_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2))
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.moop_echo, 4)
        self.assertEqual(response.service_data.max_length, 0xabcd)
        self.assertEqual(response.service_data.dfi.compression, 5)
        self.assertEqual(response.service_data.dfi.encryption, 2)
        self.assertEqual(response.service_data.filesize.uncompressed, 0x9876)
        self.assertEqual(response.service_data.filesize.compressed, 0x1234)
        self.assertIsNone(response.service_data.dirinfo_length)

    def test_read_file_extra_bytes_response_zero_padding_not_ok(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x04\x02\xab\xcd\x52\x00\x02\x98\x76\x12\x34\x00")    # Positive response

    def _test_read_file_extra_bytes_response_zero_padding_not_ok(self):
        self.udsclient.config['tolerate_zero_padding'] = False
        with self.assertRaises(InvalidResponseException):
            self.udsclient.read_file("my_file.txt", dfi= DataFormatIdentifier(compression=5, encryption=2))


    #============== Read Directory ================

    def test_read_dir_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x38\x05\x00\x0c" + "/path/to/dir".encode('ascii'))
        self.conn.fromuserqueue.put(b"\x78\x05\x02\xab\xcd\x00\x00\x02\x12\x34")    # Positive response

    def _test_read_dir_success(self):
        response = self.udsclient.read_dir("/path/to/dir")
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.moop_echo, 5)
        self.assertEqual(response.service_data.max_length, 0xabcd)
        self.assertEqual(response.service_data.dfi.encryption, 0)
        self.assertEqual(response.service_data.dfi.compression, 0)
        self.assertIsNone(response.service_data.filesize)
        self.assertEqual(response.service_data.dirinfo_length, 0x1234)

    def test_read_dir_negative_response(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x7F\x38\x22")    # Condition not correct

    def _test_read_dir_negative_response(self):
        with self.assertRaises(NegativeResponseException):
            self.udsclient.read_dir("/path/to/dir")

    def test_read_dir_negative_response_no_exception(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x7F\x38\x22")    # Condition not correct

    def _test_read_dir_negative_response_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.read_dir("/path/to/dir")
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)          

    def test_read_dir_bad_moop_echo(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x01\x02\xab\xcd\x00\x00\x02\x12\x34")    # Positive response

    def _test_read_dir_bad_moop_echo(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.read_dir("/path/to/dir")

    def test_read_dir_bad_moop_echo_no_exception(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x01\x02\xab\xcd\x00\x00\x02\x12\x34")    # Positive response

    def _test_read_dir_bad_moop_echo_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False        
        response = self.udsclient.read_dir("/path/to/dir")
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertTrue(response.unexpected)

    def test_read_dir_dfi_not_zero(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x02\xab\xcd\x52\x00\x02\x12\x34")    # Positive response

    def _test_read_dir_dfi_not_zero(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.read_dir("/path/to/dir")

    def test_read_dir_zerolen_length_lfi(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x00\x00\x00\x02\x12\x34")

    def _test_read_dir_zerolen_length_lfi(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.read_dir("/path/to/dir")

    def test_read_dir_zerolen_length_fsodipl(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x02\xab\xcd\x00\x00\x00")

    def _test_read_dir_zerolen_length_fsodipl(self):
        with self.assertRaises(InvalidResponseException):
            self.udsclient.read_dir("/path/to/dir")

    def test_read_dir_invalid_length(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x02")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x02\xab")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x02\xab\xcd")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x02\xab\xcd\x00")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x02\xab\xcd\x00\x00")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x02\xab\xcd\x00\x00\x02")
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x02\xab\xcd\x00\x00\x02\x12")

    def _test_read_dir_invalid_length(self):
        for i in range(9):
            with self.assertRaises(InvalidResponseException):
                self.udsclient.read_dir("/path/to/dir")

    def test_read_dir_extra_bytes_response(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x02\xab\xcd\x00\x00\x02\x12\x34\xAA")    # Positive response

    def _test_read_dir_extra_bytes_response(self):
        with self.assertRaises(InvalidResponseException):
            response = self.udsclient.read_dir("/path/to/dir")

    def test_read_dir_extra_bytes_response_zero_padding_ok(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x02\xab\xcd\x00\x00\x02\x12\x34\x00")    # Positive response

    def _test_read_dir_extra_bytes_response_zero_padding_ok(self):
        self.udsclient.config['tolerate_zero_padding'] = True
        response = response = self.udsclient.read_dir("/path/to/dir")
        self.assertTrue(response.valid)
        self.assertTrue(response.positive)
        self.assertEqual(response.service_data.moop_echo, 5)
        self.assertEqual(response.service_data.max_length, 0xabcd)
        self.assertEqual(response.service_data.dfi.compression, 0)
        self.assertEqual(response.service_data.dfi.encryption, 0)
        self.assertIsNone(response.service_data.filesize)
        self.assertEqual(response.service_data.dirinfo_length, 0x1234)

    def test_read_dir_extra_bytes_response_zero_padding_not_ok(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x78\x05\x02\xab\xcd\x00\x00\x02\x12\x34\x00")    # Positive response

    def _test_read_dir_extra_bytes_response_zero_padding_not_ok(self):
        self.udsclient.config['tolerate_zero_padding'] = False
        with self.assertRaises(InvalidResponseException):
            response = self.udsclient.read_dir("/path/to/dir")


    # ============= Misc ==========================

    def test_wrong_service_reply(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x71\x01\x02\xab\xcd\x00")   # Reply service ID of routine control

    def _test_wrong_service_reply(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.add_file("my_file.txt", filesize = 0x100)

    def test_wrong_service_reply_no_exception(self):
        self.conn.touserqueue.get(timeout=0.2)
        self.conn.fromuserqueue.put(b"\x71\x01\x02\xab\xcd\x00")   # Reply service ID of routine control

    def _test_wrong_service_reply_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.add_file("my_file.txt", filesize=0x100)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_default_dfi(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x38\x01\x00\x0b" + "my_file.txt".encode('ascii') + b"\x00\x04\x00\x00\x02\x22\x00\x00\x01\x11")
        self.conn.fromuserqueue.put(b"\x78\x01\x02\xab\xcd\x00")    # Positive response

    def _test_default_dfi(self):
        response = self.udsclient.request_file_transfer(moop=1, path="my_file.txt", filesize = Filesize(uncompressed=0x222, compressed=0x111, width=4))
        self.assertEqual(response.service_data.dfi.compression, 0)
        self.assertEqual(response.service_data.dfi.encryption, 0)

    def test_default_filesize_width(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x38\x01\x00\x0b" + "my_file.txt".encode('ascii') + b"\x00\x02\x02\x22\x01\x11")
        self.conn.fromuserqueue.put(b"\x78\x01\x02\xab\xcd\x00")    # Positive response

    def _test_default_filesize_width(self):
        self.udsclient.request_file_transfer(moop=1, path="my_file.txt", filesize = Filesize(uncompressed=0x222, compressed=0x111))

    def test_filesize_no_compressed_size(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x38\x01\x00\x0b" + "my_file.txt".encode('ascii') + b"\x00\x04\x00\x00\x02\x22\x00\x00\x02\x22")
        self.conn.fromuserqueue.put(b"\x78\x01\x02\xab\xcd\x00")    # Positive response

    def _test_filesize_no_compressed_size(self):
        self.udsclient.request_file_transfer(moop=1, path="my_file.txt", filesize = Filesize(uncompressed=0x222, width=4))

    def test_filesize_numerical(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x38\x01\x00\x0b" + "my_file.txt".encode('ascii') + b"\x00\x02\x01\x00\x01\x00")
        self.conn.fromuserqueue.put(b"\x78\x01\x02\xab\xcd\x00")    # Positive response

    def _test_filesize_numerical(self):
        self.udsclient.request_file_transfer(moop=1, path="my_file.txt", filesize = 0x100)

    def test_bad_param(self):
        request = None
        try:
            self.conn.touserqueue.get(timeout=0.2)
        except:
            pass
        self.assertIsNone(request)

    def _test_bad_param(self):
        good_filesize = Filesize(uncompressed=0x222, compressed=0x111, width=4)
        good_dfi = DataFormatIdentifier(compression=5, encryption=2)
        with self.assertRaises(ValueError):
            self.udsclient.request_file_transfer(moop=1, path="a"*(2**16), dfi=good_dfi, filesize = good_filesize)

        with self.assertRaises(ValueError):
            self.udsclient.request_file_transfer(moop=1, path="", dfi=good_dfi, filesize = good_filesize)

        with self.assertRaises(ValueError):
            self.udsclient.request_file_transfer(moop=1, path="hello.txt", dfi=1, filesize = good_filesize)

        with self.assertRaises(ValueError):
            self.udsclient.request_file_transfer(moop=1, path="hello.txt", dfi=good_dfi, filesize = "asd")

        for moop in [2, 4, 5]:  # Delete File, Read File, Read Dir
            with self.assertRaises(ValueError):
                self.udsclient.request_file_transfer(moop=moop, path="hello.txt", dfi=good_dfi, filesize = good_filesize)   # unexpected filesize

        for moop in [1, 3]: # Add File, Replace File
            with self.assertRaises(ValueError):
                self.udsclient.request_file_transfer(moop=moop, path="hello.txt", dfi=good_dfi, filesize = None)    # missing filesize

        for moop in [2, 5]:
            with self.assertRaises(ValueError):
                self.udsclient.request_file_transfer(moop=moop, path="hello.txt", dfi=good_dfi, filesize = None)    # dfi not expected



from test.ClientServerTest import ClientServerTest
from udsoncan import MemoryLocation
from udsoncan.exceptions import *

# Note : 
# MemoryLocation object is unit tested in a separate file (test_helper_class). 
# As it is the only parameter to be passed, no need to push this test too far for nothing.

class TestReadMemoryByAddress(ClientServerTest):

    def test_4byte_block(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x23\x12\x12\x34\x04")
        self.conn.fromuserqueue.put(b"\x63\x99\x88\x77\x66")

    def _test_4byte_block(self):
        response = self.udsclient.read_memory_by_address(MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8))
        self.assertEqual(response.service_data.memory_block, b'\x99\x88\x77\x66')

    def test_4byte_block_spr_no_effect(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x23\x12\x12\x34\x04")
        self.conn.fromuserqueue.put(b"\x63\x99\x88\x77\x66")

    def _test_4byte_block_spr_no_effect(self):
        with self.udsclient.suppress_positive_response:
            response = self.udsclient.read_memory_by_address(MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8))
            self.assertEqual(response.service_data.memory_block, b'\x99\x88\x77\x66')		

    def test_config_format(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x23\x24\x00\x00\x12\x34\x00\x04")
        self.conn.fromuserqueue.put(b"\x63\x99\x88\x77\x66")

    def _test_config_format(self):
        self.udsclient.config['server_address_format'] = 32
        self.udsclient.config['server_memorysize_format'] = 16
        self.udsclient.read_memory_by_address(MemoryLocation(address=0x1234, memorysize=4))

    def test_4byte_block_zeropadding_ok(self):
        data = b"\x63\x99\x88\x77\x66"
        for i in range(8):
            self.wait_request_and_respond(data + b'\x00'*(i+1))

    def _test_4byte_block_zeropadding_ok(self):
        self.udsclient.config['tolerate_zero_padding'] = True
        for i in range(8):
            response = self.udsclient.read_memory_by_address(MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8))
            self.assertEqual(response.service_data.memory_block, b'\x99\x88\x77\x66')

    def test_4byte_block_zeropadding_not_ok_exception(self):
        data = b"\x63\x99\x88\x77\x66"
        for i in range(8):
            self.wait_request_and_respond(data + b'\x00'*(i+1))

    def _test_4byte_block_zeropadding_not_ok_exception(self):
        self.udsclient.config['tolerate_zero_padding'] = False
        for i in range(8):
            with self.assertRaises(UnexpectedResponseException):
                self.udsclient.read_memory_by_address(MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8))

    def test_4byte_block_zeropadding_not_ok_no_exception(self):
        data = b"\x63\x99\x88\x77\x66"
        for i in range(8):
            self.wait_request_and_respond(data + b'\x00'*(i+1))

    def _test_4byte_block_zeropadding_not_ok_no_exception(self):
        self.udsclient.config['tolerate_zero_padding'] = False
        self.udsclient.config['exception_on_unexpected_response'] = False
        for i in range(8):
            response = self.udsclient.read_memory_by_address(MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8))
            self.assertTrue(response.valid)
            self.assertTrue(response.unexpected)

    def test_request_denied_exception(self):
        self.wait_request_and_respond(b"\x7F\x23\x45") #Request Out Of Range

    def _test_request_denied_exception(self):
        with self.assertRaises(NegativeResponseException) as handle:
            self.udsclient.read_memory_by_address(MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8))

    def test_request_denied_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x23\x45") #Request Out Of Range

    def _test_request_denied_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.read_memory_by_address(MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8))
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)

    def test_request_invalid_service_exception(self):
        self.wait_request_and_respond(b"\x00\x45") #Inexistent Service

    def _test_request_invalid_service_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.read_memory_by_address(MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8))

    def test_request_invalid_service_no_exception(self):
        self.wait_request_and_respond(b"\x00\x45") #Inexistent Service

    def _test_request_invalid_service_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.read_memory_by_address(MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8))
        self.assertFalse(response.valid)

    def test_wrong_service_exception(self):
        self.wait_request_and_respond(b"\x7E\x99\x88\x77\x66") # Valid but wrong service (Tester Present)

    def _test_wrong_service_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.read_memory_by_address(MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8))

    def test_wrong_service_no_exception(self):
        self.wait_request_and_respond(b"\x7E\x99\x88\x77\x66") # Valid but wrong service (Tester Present)

    def _test_wrong_service_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.read_memory_by_address(MemoryLocation(address=0x1234, memorysize=4, address_format=16, memorysize_format=8))
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_bad_param(self):
        pass

    def _test_bad_param(self):
        with self.assertRaises(ValueError):
            self.udsclient.read_memory_by_address(1)

        with self.assertRaises(ValueError):
            self.udsclient.read_memory_by_address('aaa')



__all__ = ['CommunicationType']

# Communication type is a single byte value including message type and subnet.
# Used by CommunicationControl service and defined by ISO-14229:2006 Annex B, table B.1
import struct
from udsoncan import tools

from typing import Union, List


class CommunicationType:
    """
    This class represents a pair of subnet and message types. This value is mainly used by the :ref:`CommunicationControl<CommunicationControl>` service

    :param subnet: Represent the subnet number. Value ranges from 0 to 0xF 
    :type subnet: int

    :param normal_msg: Bit indicating that the `normal messages` are involved
    :type normal_msg: bool

    :param network_management_msg: Bit indicating that the `network management messages` are involved
    :type network_management_msg: bool

    """
    class Subnet:
        node = 0
        network = 0xF

        def __init__(self, subnet: int):
            tools.validate_int(subnet, min=0, max=0xF, name="subnet")
            self.subnet = subnet

        def value(self):
            return self.subnet

    subnet: Subnet
    normal_msg: bool
    network_management_msg: bool

    def __init__(self, subnet: Union[Subnet, int], normal_msg: bool = False, network_management_msg: bool = False):

        if not isinstance(subnet, self.Subnet):
            subnet = self.Subnet(subnet)

        if not isinstance(normal_msg, bool) or not isinstance(network_management_msg, bool):
            raise ValueError('message type (normal_msg, network_management_msg) must be valid boolean values')

        if normal_msg == False and network_management_msg == False:
            raise ValueError('At least one message type must be controlled')

        self.subnet = subnet
        self.normal_msg = normal_msg
        self.network_management_msg = network_management_msg

    def get_byte_as_int(self) -> int:
        message_type = 0
        if self.normal_msg:
            message_type |= 1
        if self.network_management_msg:
            message_type |= 2

        byte = (message_type & 0x3) | ((self.subnet.value() & 0xF) << 4)
        return byte

    def get_byte(self):
        return struct.pack('B', self.get_byte_as_int())

    @classmethod
    def from_byte(cls, val: Union[int, bytes]) -> "CommunicationType":
        if isinstance(val, bytes):
            val = struct.unpack('B', val)[0]
        val = int(val)
        subnet = (val & 0xF0) >> 4
        normal_msg = True if val & 1 > 0 else False
        network_management_msg = True if val & 2 > 0 else False
        return cls(subnet, normal_msg, network_management_msg)

    def __str__(self) -> str:
        flags: List[str] = []
        if self.normal_msg:
            flags.append('NormalMsg')

        if self.network_management_msg:
            flags.append('NetworkManagementMsg')

        return 'subnet=0x%x. Flags : [%s]' % (self.subnet.value(), ','.join(flags))

    def __repr__(self) -> str:
        return '<%s: %s at 0x%08x>' % (self.__class__.__name__, str(self), id(self))



__all__ = [
    'DidCodec',
    'AsciiCodec'
]

import struct

from typing import Optional, Any


class DidCodec:
    """
    This class defines how to encode/decode a Data Identifier value to/from a binary payload.

    One should extend this class and override the ``encode``, ``decode``, ``__len__`` methods as they will be used
    to generate or parse binary payloads.

            - ``encode`` Must receive any Python object and must return a bytes payload
            - ``decode`` Must receive a bytes payload and may return any Python object
            - ``__len__`` Must return the length of the bytes payload or raise a ``DidCodec.ReadAllRemainingData`` to read the whole payload. Reading the whole payload is not a feature proposed by ISO-14229.

    If a data can be processed by a pack string, then this class may be used as is, without being extended.

    :param packstr: A pack string used with struct.pack / struct.unpack. 
    :type packstr: string
    """

    class ReadAllRemainingData(Exception):
        pass

    packstr: Optional[str]

    def __init__(self, packstr: Optional[str] = None):
        self.packstr = packstr

    def encode(self, *did_value: Any) -> bytes:
        if self.packstr is None:
            raise NotImplementedError('Cannot encode DID to binary payload. Codec has no "encode" implementation')

        return struct.pack(self.packstr, *did_value)

    def decode(self, did_payload: bytes) -> Any:
        if self.packstr is None:
            raise NotImplementedError('Cannot decode DID from binary payload. Codec has no "decode" implementation')

        return struct.unpack(self.packstr, did_payload)

    # Must tell the size of the payload encoded or expected for decoding
    def __len__(self) -> int:
        if self.packstr is None:
            raise NotImplementedError('Cannot tell the payload size. Codec has no "__len__" implementation')
        return struct.calcsize(self.packstr)


class AsciiCodec(DidCodec):
    string_len: int

    def __init__(self, string_len: int):
        self.string_len = string_len

    def encode(self, string_ascii: Any) -> bytes:
        if not isinstance(string_ascii, str):
            raise ValueError("AsciiCodec requires a string for encoding")

        if len(string_ascii) != self.string_len:
            raise ValueError('String must be %d long' % self.string_len)
        return string_ascii.encode('ascii')

    def decode(self, string_bin: bytes) -> Any:
        string_ascii = string_bin.decode('ascii')
        if len(string_ascii) != self.string_len:
            raise ValueError('Trying to decode a string of %d bytes but codec expects %d bytes' % (len(string_ascii), self.string_len))
        return string_ascii

    def __len__(self) -> int:
        return self.string_len



import ctypes
from ctypes import Structure, WINFUNCTYPE, POINTER, cast, c_long, c_void_p, c_ulong, byref  # type: ignore

from enum import Enum

import logging


class PASSTHRU_MSG(Structure):
    _fields_ = [("ProtocolID", c_ulong),
                ("RxStatus", c_ulong),
                ("TxFlags", c_ulong),
                ("Timestamp", c_ulong),
                ("DataSize", c_ulong),
                ("ExtraDataindex", c_ulong),
                ("Data", ctypes.c_ubyte * 4128)]


class SCONFIG(Structure):
    _fields_ = [("Parameter", c_ulong),
                ("Value", c_ulong)]


class SCONFIG_LIST(Structure):
    _fields_ = [("NumOfParams", c_ulong),
                ("ConfigPtr", POINTER(SCONFIG))]


class J2534():
    dllPassThruOpen = None
    dllPassThruClose = None
    dllPassThruConnect = None
    dllPassThruDisconnect = None
    dllPassThruReadMsgs = None
    dllPassThruWriteMsgs = None
    dllPassThruStartPeriodicMsg = None
    dllPassThruStopPeriodicMsg = None
    dllPassThruReadVersion = None
    dllPassThruStartMsgFilter = None
    dllPassThruIoctl = None

    def __init__(self, windll, rxid, txid):

        global dllPassThruOpen
        global dllPassThruClose
        global dllPassThruConnect
        global dllPassThruDisconnect
        global dllPassThruReadMsgs
        global dllPassThruWriteMsgs
        global dllPassThruStartPeriodicMsg
        global dllPassThruStopPeriodicMsg
        global dllPassThruReadVersion
        global dllPassThruStartMsgFilter
        global dllPassThruIoctl

        self.hDLL = ctypes.cdll.LoadLibrary(windll)
        self.rxid = rxid.to_bytes(4, 'big')
        self.txid = txid.to_bytes(4, 'big')

        self.logger = logging.getLogger()

        dllPassThruOpenProto = WINFUNCTYPE(
            c_long,
            c_void_p,
            POINTER(c_ulong))

        dllPassThruOpenParams = (1, "pName", 0), (1, "pDeviceID", 0)
        dllPassThruOpen = dllPassThruOpenProto(("PassThruOpen", self.hDLL), dllPassThruOpenParams)

        dllPassThruCloseProto = WINFUNCTYPE(
            c_long,
            c_ulong)

        dllPassThruCloseParams = (1, "DeviceID", 0),
        dllPassThruClose = dllPassThruCloseProto(("PassThruClose", self.hDLL), dllPassThruCloseParams)

        dllPassThruConnectProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            c_ulong,
            c_ulong,
            c_ulong,
            POINTER(c_ulong))

        dllPassThruConnectParams = (1, "DeviceID", 0), (1, "ProtocolID", 0), (1, "Flags", 0), (1, "BaudRate", 500000), (1, "pChannelID", 0)
        dllPassThruConnect = dllPassThruConnectProto(("PassThruConnect", self.hDLL), dllPassThruConnectParams)

        dllPassThruDisconnectProto = WINFUNCTYPE(
            c_long,
            c_ulong)

        dllPassThruDisconnectParams = (1, "ChannelID", 0),
        dllPassThruDisconnect = dllPassThruDisconnectProto(("PassThruDisconnect", self.hDLL), dllPassThruDisconnectParams)

        dllPassThruReadMsgsProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            POINTER(PASSTHRU_MSG),
            POINTER(c_ulong),
            c_ulong)

        dllPassThruReadMsgsParams = (1, "ChannelID", 0), (1, "pMsg", 0), (1, "pNumMsgs", 0), (1, "Timeout", 0)
        dllPassThruReadMsgs = dllPassThruReadMsgsProto(("PassThruReadMsgs", self.hDLL), dllPassThruReadMsgsParams)

        dllPassThruWriteMsgsProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            POINTER(PASSTHRU_MSG),
            POINTER(c_ulong),
            c_ulong)

        dllPassThruWriteMsgsParams = (1, "ChannelID", 0), (1, "pMsg", 0), (1, "pNumMsgs", 0), (1, "Timeout", 0)
        dllPassThruWriteMsgs = dllPassThruWriteMsgsProto(("PassThruWriteMsgs", self.hDLL), dllPassThruWriteMsgsParams)

        dllPassThruStartPeriodicMsgProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            POINTER(PASSTHRU_MSG),
            POINTER(c_ulong),
            c_ulong)

        dllPassThruStartPeriodicMsgParams = (1, "ChannelID", 0), (1, "pMsg", 0), (1, "pMsgID", 0), (1, "TimeInterval", 0)
        dllPassThruStartPeriodicMsg = dllPassThruStartPeriodicMsgProto(("PassThruStartPeriodicMsg", self.hDLL), dllPassThruStartPeriodicMsgParams)

        dllPassThruStopPeriodicMsgProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            c_ulong)

        dllPassThruStopPeriodicMsgParams = (1, "ChannelID", 0), (1, "MsgID", 0)
        dllPassThruStopPeriodicMsg = dllPassThruStopPeriodicMsgProto(("PassThruStopPeriodicMsg", self.hDLL), dllPassThruStopPeriodicMsgParams)

        dllPassThruReadVersionProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            POINTER(ctypes.c_char),
            POINTER(ctypes.c_char),
            POINTER(ctypes.c_char))

        dllPassThruReadVersionParams = (1, "DeviceID", 0), (1, "pFirmwareVersion", 0), (1, "pDllVersion", 0), (1, "pApiVersoin", 0)
        dllPassThruReadVersion = dllPassThruReadVersionProto(("PassThruReadVersion", self.hDLL), dllPassThruReadVersionParams)

        dllPassThruStartMsgFilterProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            c_ulong,
            POINTER(PASSTHRU_MSG),
            POINTER(PASSTHRU_MSG),
            POINTER(PASSTHRU_MSG),
            POINTER(c_ulong)
        )

        dllPassThruStartMsgFilterParams = (1, "ChannelID", 0), (1, "FilterType", 0), (1, "pMaskMsg", 0), (1,
                                                                                                          "pPatternMsg", 0), (1, "pFlowControlMsg", 0), (1, "pMsgID", 0)

        dllPassThruStartMsgFilter = dllPassThruStartMsgFilterProto(("PassThruStartMsgFilter", self.hDLL), dllPassThruStartMsgFilterParams)

        dllPassThruIoctlProto = WINFUNCTYPE(
            c_long,
            c_ulong,
            c_ulong,
            c_void_p,
            c_void_p
        )

        dllPassThruIoctlParams = (1, "Handle", 0), (1, "IoctlID", 0), (1, "pInput", 0), (1, "pOutput", 0)

        dllPassThruIoctl = dllPassThruIoctlProto(("PassThruIoctl", self.hDLL), dllPassThruIoctlParams)

    def PassThruOpen(self, pDeviceID=None):
        if not pDeviceID:
            pDeviceID = ctypes.c_ulong()

        result = dllPassThruOpen(ctypes.POINTER(ctypes.c_int)(), byref(pDeviceID))
        return Error_ID(hex(result)), pDeviceID

    def PassThruConnect(self, deviceID, protocol, baudrate, pChannelID=None):
        if not pChannelID:
            pChannelID = c_ulong()

        result = dllPassThruConnect(deviceID, protocol, 0, baudrate, byref(pChannelID))
        return Error_ID(hex(result)), pChannelID

    def PassThruClose(self, DeviceID):
        result = dllPassThruClose(DeviceID)
        return Error_ID(hex(result))

    def PassThruDisconnect(self, ChannelID):
        result = dllPassThruDisconnect(ChannelID)
        return Error_ID(hex(result))

    def PassThruReadMsgs(self, ChannelID, protocol, pNumMsgs=1, Timeout=100):
        pMsg = PASSTHRU_MSG()
        pMsg.ProtocolID = protocol

        pNumMsgs = c_ulong(pNumMsgs)

        while 1:
            # breakpoint()
            result = dllPassThruReadMsgs(ChannelID, byref(pMsg), byref(pNumMsgs), c_ulong(Timeout))
            if Error_ID(hex(result)) == Error_ID.ERR_BUFFER_EMPTY or pNumMsgs == 0:
                return None, None, 0
            elif pMsg.RxStatus == 0:
                return Error_ID(hex(result)), bytes(pMsg.Data[4:pMsg.DataSize]), pNumMsgs

    def PassThruWriteMsgs(self, ChannelID, Data, protocol, pNumMsgs=1, Timeout=1000):
        txmsg = PASSTHRU_MSG()
        txmsg.TxFlags = TxStatusFlag.ISO15765_FRAME_PAD.value
        txmsg.ProtocolID = protocol;

        Data = self.txid + Data
        self.logger.info("Sending data: " + str(Data.hex()))

        for i in range(0, len(Data)):
            txmsg.Data[i] = Data[i]

        txmsg.DataSize = len(Data)

        result = dllPassThruWriteMsgs(ChannelID, byref(txmsg), byref(c_ulong(pNumMsgs)), c_ulong(Timeout))

        return Error_ID(hex(result))

    def PassThruStartPeriodicMsg(self, ChannelID, Data, MsgID=0, TimeInterval=100):
        pMsg = PASSTHRU_MSG()

        pMsg.Data = Data
        pMsg.DataSize = len(Data)

        result = dllPassThruStartPeriodicMsg(ChannelID, byref(pMsg), byref(c_ulong(MsgID)), c_ulong(TimeInterval))

        return Error_ID(hex(result))

    def PassThruStopPeriodicMsg(self, ChannelID, MsgID):
        result = dllPassThruStopPeriodicMsg(ChannelID, MsgID)

        return Error_ID(hex(result))

    def PassThruReadVersion(self, DeviceID):
        pFirmwareVersion = (ctypes.c_char * 80)()
        pDllVersion = (ctypes.c_char * 80)()
        pApiVersion = (ctypes.c_char * 80)()
        result = dllPassThruReadVersion(DeviceID, pFirmwareVersion, pDllVersion, pApiVersion)

        return Error_ID(hex(result)), pFirmwareVersion, pDllVersion, pApiVersion

    def PassThruIoctl(self, Handle, IoctlID, ioctlInput=None, ioctlOutput=None):

        if ioctlInput is None:
            pInput = POINTER(c_ulong)()
        else:
            pInput = c_ulong(ioctlInput.value)

        if ioctlOutput is None:
            pOutput = POINTER(c_ulong)()

        result = dllPassThruIoctl(Handle, c_ulong(IoctlID.value), byref(pInput), byref(pOutput))

        return Error_ID(hex(result))

    def PassThruStartMsgFilter(self, ChannelID, protocol):

        txmsg = PASSTHRU_MSG()
        msgMask = PASSTHRU_MSG()
        msgPattern = PASSTHRU_MSG()
        msgFlow = PASSTHRU_MSG()

        txmsg.ProtocolID = protocol;
        txmsg.RxStatus = 0;
        txmsg.TxFlags = TxStatusFlag.ISO15765_FRAME_PAD.value
        txmsg.Timestamp = 0;
        txmsg.DataSize = 4;

        msgMask.ProtocolID = protocol;
        msgMask.RxStatus = 0;
        msgMask.TxFlags = TxStatusFlag.ISO15765_FRAME_PAD.value
        msgMask.Timestamp = 0;
        msgMask.DataSize = 4;
        for i in range(0, 4):
            msgMask.Data[i] = 0xff

        msgPattern.ProtocolID = protocol;
        msgPattern.RxStatus = 0;
        msgPattern.TxFlags = TxStatusFlag.ISO15765_FRAME_PAD.value
        msgPattern.Timestamp = 0;
        msgPattern.DataSize = 4;

        msgFlow.ProtocolID = protocol;
        msgFlow.RxStatus = 0;
        msgFlow.TxFlags = TxStatusFlag.ISO15765_FRAME_PAD.value
        msgFlow.Timestamp = 0;
        msgFlow.DataSize = 4;

        for i in range(0, len(self.txid)):
            msgFlow.Data[i] = self.txid[i]

        for i in range(0, len(self.rxid)):
            msgPattern.Data[i] = self.rxid[i]

        msgID = c_ulong(0)

        result = dllPassThruStartMsgFilter(ChannelID, c_ulong(Filter.FLOW_CONTROL_FILTER.value),
                                           byref(msgMask), byref(msgPattern), byref(msgFlow), byref(msgID))

        # for i in range(0, len(self.rxid)):
        #    msgFlow.Data[i] = self.rxid[i]

        # for i in range(0, len(self.txid)):
        #    msgPattern.Data[i] = self.txid[i]

        result = dllPassThruStartMsgFilter(ChannelID, c_ulong(Filter.FLOW_CONTROL_FILTER.value),
                                           byref(msgMask), byref(msgPattern), byref(msgFlow), byref(msgID))

        return Error_ID(hex(result))


class Error_ID(Enum):

    ERR_SUCCESS = hex(0x00)
    STATUS_NOERROR = hex(0x00)
    ERR_NOT_SUPPORTED = hex(0x01)
    ERR_INVALID_CHANNEL_ID = hex(0x02)
    ERR_INVALID_PROTOCOL_ID = hex(0x03)
    ERR_NULL_PARAMETER = hex(0x04)
    ERR_INVALID_IOCTL_VALUE = hex(0x05)
    ERR_INVALID_FLAGS = hex(0x06)
    ERR_FAILED = hex(0x07)
    ERR_DEVICE_NOT_CONNECTED = hex(0x08)
    ERR_TIMEOUT = hex(0x09)
    ERR_INVALID_MSG = hex(0x0A)
    ERR_INVALID_TIME_INTERVAL = hex(0x0B)
    ERR_EXCEEDED_LIMIT = hex(0x0C)
    ERR_INVALID_MSG_ID = hex(0x0D)
    ERR_DEVICE_IN_USE = hex(0x0E)
    ERR_INVALID_IOCTL_ID = hex(0x0F)
    ERR_BUFFER_EMPTY = hex(0x10)
    ERR_BUFFER_FULL = hex(0x11)
    ERR_BUFFER_OVERFLOW = hex(0x12)
    ERR_PIN_INVALID = hex(0x13)
    ERR_CHANNEL_IN_USE = hex(0x14)
    ERR_MSG_PROTOCOL_ID = hex(0x15)
    ERR_INVALID_FILTER_ID = hex(0x16)
    ERR_NO_FLOW_CONTROL = hex(0x17)
    ERR_NOT_UNIQUE = hex(0x18)
    ERR_INVALID_BAUDRATE = hex(0x19)
    ERR_INVALID_DEVICE_ID = hex(0x1A)


class Protocol_ID(Enum):

    J1850VPW = 1
    J1850PWM = 2
    ISO9141 = 3
    ISO14230 = 4
    CAN = 5
    ISO15765 = 6
    SCI_A_ENGINE = 7  # OP2.0: Not supported
    SCI_A_TRANS = 8  # OP2.0: Not supported
    SCI_B_ENGINE = 9  # OP2.0: Not supported
    SCI_B_TRANS = 10  # OP2.0: Not supported


class Filter(Enum):
    PASS_FILTER = 0x00000001
    BLOCK_FILTER = 0x00000002
    FLOW_CONTROL_FILTER = 0x00000003


class TxStatusFlag(Enum):
    ISO15765_FRAME_PAD = 0x00000040
    WAIT_P3_MIN_ONLY = 0x00000200
    SW_CAN_HV_TX = 0x00000400  # OP2.0: Not supported
    SCI_MODE = 0x00400000  # OP2.0: Not supported
    SCI_TX_VOLTAGE = 0x00800000  # OP2.0: Not supported


class Ioctl_ID(Enum):
    GET_CONFIG = 0x01
    SET_CONFIG = 0x02
    CLEAR_RX_BUFFER = 0x08


class Ioctl_Flags(Enum):
    TX_IOCTL_BASE = 0x70000
    TX_IOCTL_SET_DLL_DEBUG_FLAGS = 0x70001
    TX_IOCTL_DLL_DEBUG_FLAG_J2534_CALLS = 0x00000001



from udsoncan.BaseService import BaseService, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
import inspect
import struct

from typing import Type, Optional, Union


class Response:
    """
    Represents a server Response to a client Request

    :param service: The service implied by this response.
    :type service: class

    :param code: The response code
    :type code: int

    :param data: The response data encoded after the service and response code
    :type data: bytes

    .. data:: valid 

            (boolean) True if the response content is valid. Only ``invalid_reason`` is guaranteed to have a meaningful value if this value is False

    .. data:: invalid_reason 

            (string) String explaining why the response is invalid.

    .. data:: service 

            (class) The response target :ref:`service<Services>` class

    .. data:: positive 

            (boolean) True if the response code is 0 (PositiveResponse), False otherwise

    .. data:: code 

            (int) The response code. 

    .. data:: code_name 

            (string) The response code name.


    .. data:: data

            (bytes) The response data. All the payload content, except the service number and the response code


    .. data:: service_data

            (object) The content of ``data`` interpreted by a service; can be any type of content.


    .. data:: original_payload 

            (bytes) When the response is built with `Response.from_payload`, this property contains a copy of the payload used. None otherwise.

    .. data:: unexpected 

            (boolean) Indicates that the response was unexpected. Set by an external source such as the :ref:`Client<Client>` object

    """
    Code = ResponseCode

    service: Optional[Type[BaseService]]
    subfunction: Optional[int]
    data: Optional[bytes]
    suppress_positive_response: bool
    original_payload: Optional[bytes]
    service_data: Optional[BaseResponseData]

    def __init__(self,
                 service: Optional[Union[BaseService, Type[BaseService]]] = None,
                 code: Optional[int] = None,
                 data: Optional[bytes] = None):
        if service is None:
            self.service = None
        elif isinstance(service, BaseService):
            self.service = service.__class__
        elif inspect.isclass(service) and issubclass(service, BaseService):
            self.service = service
        elif service is not None:
            raise ValueError("Given service must be a service class or instance")

        self.positive = False
        self.code = None
        self.code_name = ""
        self.valid = False
        self.invalid_reason = "Object not initialized"
        self.service_data = None
        self.original_payload = None
        self.unexpected = False

        if data is not None:
            if not isinstance(data, bytes):
                raise ValueError("Given data must be a valid bytes object")

        self.data = data if data is not None else b''

        if code is not None:
            if not isinstance(code, int):
                raise ValueError("Response code must be a valid integer")
            elif code < 0 or code > 0xFF:
                raise ValueError("Response code must be an integer between 0 and 0xFF")
            self.code = code
            self.code_name = Response.Code.get_name(code)
            if not Response.Code.is_negative(code):
                self.positive = True

        if self.service is not None and self.code is not None:
            self.valid = True
            self.invalid_reason = ""

    # Used by server
    def get_payload(self) -> bytes:
        """
        Generates a payload to be given to the underlying protocol.
        This method is meant to be used by a UDS server

        :return: A payload to be sent through the underlying protocol
        :rtype: bytes
        """
        if self.service is None:
            raise ValueError("Cannot make payload from response object. Service is not set")

        if not isinstance(self.service, BaseService) and not issubclass(self.service, BaseService):
            raise ValueError("Cannot make payload from response object. Given service is not a valid service object")

        if not isinstance(self.code, int):
            raise ValueError("Cannot make payload from response object. Given response code is not a valid integer")

        payload = b''
        if self.positive:
            payload += struct.pack("B", self.service.response_id())
        else:
            payload += b'\x7F'
            payload += struct.pack("B", self.service.request_id())
            payload += struct.pack('B', self.code)

        if self.data is not None and self.service.has_response_data():
            payload += self.data
        return payload

    # Analyzes a TP frame and builds a Response object. Used by client

    @classmethod
    def from_payload(cls, payload: bytes) -> "Response":
        """
        Creates a ``Response`` object from a payload coming from the underlying protocol.
        This method is meant to be used by a UDS client

        :param payload: The payload of data to parse
        :type payload: bytes

        :return: A :ref:`Response<Response>` object with populated fields
        :rtype: :ref:`Response<Response>`
        """
        response = cls()
        response.original_payload = payload  # may be useful for debugging

        if len(payload) < 1:
            response.valid = False
            response.invalid_reason = "Payload is empty"
            return response

        if payload[0] != 0x7F:  # Positive
            response.service = BaseService.from_response_id(payload[0])
            if response.service is None:
                response.valid = False
                response.invalid_reason = "Payload first byte is not a know service response ID."
                return response

            data_start = 1
            response.positive = True
            if len(payload) < 2 and response.service.has_response_data():
                response.valid = False
                response.positive = False
                response.invalid_reason = "Payload must be at least 2 bytes long (service and response)"
                return response

            response.code = Response.Code.PositiveResponse
            response.code_name = Response.Code.get_name(Response.Code.PositiveResponse)

        else:  # Negative response
            response.positive = False
            data_start = 3

            if len(payload) < 2:
                response.valid = False
                response.invalid_reason = "Incomplete invalid response service (7Fxx)"
                return response
            response.service = BaseService.from_request_id(payload[1])  # Request id, not response id

            if response.service is None:
                response.valid = False
                response.invalid_reason = "Payload second byte is not a known service request ID."
                return response

            if len(payload) < 3:
                response.valid = False
                response.invalid_reason = "Response code missing"
                return response

            response.code = int(payload[2])
            response.code_name = Response.Code.get_name(response.code)

        response.valid = True
        response.invalid_reason = ""
        if len(payload) > data_start:
            response.data = payload[data_start:]
        return response

    def __repr__(self) -> str:
        responsename = Response.Code.get_name(Response.Code.PositiveResponse) if self.positive else 'NegativeResponse(%s)' % self.code_name
        bytesize = len(self.data) if self.data is not None else 0
        service_name = "NoService"
        if self.service is not None:
            service_name = self.service.get_name()
        return '<%s: [%s] - %d data bytes at 0x%08x>' % (responsename, service_name, bytesize, id(self))

    def __len__(self) -> int:
        try:
            return len(self.get_payload())
        except:
            return 0



from udsoncan.BaseService import BaseService
import udsoncan.tools as tools
import inspect
import struct

from typing import Type, Optional, Union


class Request:
    """
    Represents a UDS Request.

    :param service: The service for which to make the request. This parameter must be a class that extends :class:`udsoncan.services.BaseService`
    :type service: class

    :param subfunction: The service subfunction. This value may be ignored if the given service does not supports subfunctions
    :type subfunction: int or None

    :param suppress_positive_response: Indicates that the server should not send a response if the response code is positive. 
            This parameter has effect only when the given service supports subfunctions
    :type suppress_positive_response: bool

    :param data: The service data appended after service ID and payload
    :type data: bytes
    """
    service: Optional[Type[BaseService]]
    subfunction: Optional[int]
    data: Optional[bytes]
    suppress_positive_response: bool

    def __init__(self,
                 service: Optional[Union[BaseService, Type[BaseService]]] = None,
                 subfunction: Optional[int] = None,
                 suppress_positive_response: bool = False,
                 data: Optional[bytes] = None):
        if service is None:
            self.service = None
        elif isinstance(service, BaseService):
            self.service = service.__class__
        elif inspect.isclass(service) and issubclass(service, BaseService):
            self.service = service
        elif service is not None:
            raise ValueError("Given service must be a service class or instance")

        if not isinstance(suppress_positive_response, bool):
            raise ValueError("suppress_positive_response must be a boolean value")

        if subfunction is not None:
            if isinstance(subfunction, int):
                self.subfunction = subfunction
            else:
                raise ValueError("Given subfunction must be a valid integer")
        else:
            self.subfunction = None

        if self.service is not None:
            if suppress_positive_response and self.service.use_subfunction() == False:
                raise ValueError('Cannot suppress positive response for service %s. This service does not have a subfunction' %
                                 (self.service.get_name()))
        self.suppress_positive_response = suppress_positive_response

        if data is not None and not isinstance(data, bytes):
            raise ValueError("data must be a valid bytes object")

        self.data = data

    def get_payload(self, suppress_positive_response: Optional[bool] = None) -> bytes:
        """
        Generates a payload to be given to the underlying protocol.
        This method is meant to be used by a UDS client

        :return: A payload to be sent through the underlying protocol
        :rtype: bytes
        """
        if self.service is None:
            raise ValueError("Cannot generate a payload. Service not set")

        if not issubclass(self.service, BaseService):
            raise ValueError("Cannot generate a payload. Given service is not a subclass of BaseService")
        use_subfunction = self.service.use_subfunction()
        if use_subfunction and not isinstance(self.subfunction, int):
            raise ValueError("Cannot generate a payload. Given subfunction is not a valid integer")

        requestid = self.service.request_id()  # Returns the service ID used to make a client request

        payload = struct.pack("B", requestid)
        if use_subfunction:
            assert self.subfunction is not None
            subfunction = self.subfunction
            if suppress_positive_response is None:
                if self.suppress_positive_response:
                    subfunction |= 0x80
            else:
                if suppress_positive_response == True:
                    subfunction |= 0x80
                elif suppress_positive_response == False:
                    subfunction &= ~0x80
            payload += struct.pack("B", subfunction)
        else:
            if suppress_positive_response == True or self.suppress_positive_response == True:
                raise ValueError('Cannot suppress positive response for service %s. This service does not have a subfunction' %
                                 (self.service.get_name()))

        if self.data is not None:
            payload += self.data

        return payload

    @classmethod
    def from_payload(cls, payload: bytes) -> "Request":
        """
        Creates a ``Request`` object from a payload coming from the underlying protocols.
        This method is meant to be used by a UDS server

        :param payload: The payload of data to parse
        :type payload: bytes

        :return: A :ref:`Request<Request>` object with populated fields
        :rtype: :ref:`Request<Request>`
        """
        req = cls()

        if len(payload) >= 1:
            req.service = tools.cls_from_request_id(payload[0])
            if req.service is not None:		# Invalid service ID will make service None
                offset = 0
                if req.service.use_subfunction():
                    offset += 1
                    if len(payload) >= offset + 1:
                        req.subfunction = int(payload[1]) & 0x7F
                        req.suppress_positive_response = True if payload[1] & 0x80 > 0 else False
                if len(payload) > offset + 1:
                    req.data = payload[offset + 1:]
        return req

    def __repr__(self) -> str:
        suppress_positive_response = '[SuppressPosResponse] ' if self.suppress_positive_response else ''
        service_name = 'NoService'
        if self.service is not None:
            service_name = self.service.get_name()
        subfunction_name = ''
        if self.subfunction is not None and (self.service is None or self.service is not None and self.service.use_subfunction()):
            subfunction_name = '(subfunction=%d) ' % self.subfunction
        bytesize = len(self.data) if self.data is not None else 0
        return '<Request: [%s] %s- %d data bytes %sat 0x%08x>' % (service_name, subfunction_name, bytesize, suppress_positive_response, id(self))

    def __len__(self) -> int:
        try:
            return len(self.get_payload())
        except:
            return 0



from udsoncan.client import Client
from udsoncan import services
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest

class TestRequestTransferExit(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def test_request_transfer_exit_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x37\x12\x34\x56")
        self.conn.fromuserqueue.put(b"\x77\x89\xab\xcd\xef")	# Positive response

    def _test_request_transfer_exit_success(self):
        response = self.udsclient.request_transfer_exit(data=b'\x12\x34\x56')
        self.assertEqual(response.service, services.RequestTransferExit)
        self.assertEqual(response.service_data.parameter_records, b'\x89\xab\xcd\xef')

    def test_request_transfer_exit_success_spr_no_effect(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x37\x12\x34\x56")
        self.conn.fromuserqueue.put(b"\x77\x89\xab\xcd\xef")	# Positive response

    def _test_request_transfer_exit_success_spr_no_effect(self):
        with self.udsclient.suppress_positive_response:
            response = self.udsclient.request_transfer_exit(data=b'\x12\x34\x56')
            self.assertEqual(response.service, services.RequestTransferExit)
            self.assertEqual(response.service_data.parameter_records, b'\x89\xab\xcd\xef')

    def test_request_transfer_exit_no_data_ok(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x37")
        self.conn.fromuserqueue.put(b"\x77")	# Positive response

    def _test_request_transfer_exit_no_data_ok(self):
        response = self.udsclient.request_transfer_exit()
        self.assertEqual(response.service_data.parameter_records, b'')	

    def test_request_transfer_exit_denied_exception(self):
        self.wait_request_and_respond(b"\x7F\x37\x24") # reset sequence error

    def _test_request_transfer_exit_denied_exception(self):
        with self.assertRaises(NegativeResponseException) as handle:
            self.udsclient.request_transfer_exit()
        response = handle.exception.response

        self.assertTrue(response.valid)
        self.assertTrue(issubclass(response.service, services.RequestTransferExit))
        self.assertEqual(response.code, 0x24)

    def test_request_transfer_exit_denied_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x37\x24") # reset sequence error

    def _test_request_transfer_exit_denied_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False	
        response = self.udsclient.request_transfer_exit()
        self.assertTrue(response.valid)
        self.assertFalse(response.positive)

    def test_request_transfer_exit_invalidservice_exception(self):
        self.wait_request_and_respond(b"\x00\x55") #Inexistent Service

    def _test_request_transfer_exit_invalidservice_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.request_transfer_exit(data=b'\x12\x34\x56')

    def test_request_transfer_exit_invalidservice_no_exception(self):
        self.wait_request_and_respond(b"\x00\x55") #Inexistent Service

    def _test_request_transfer_exit_invalidservice_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False	
        response = self.udsclient.request_transfer_exit(data=b'\x12\x34\x56')
        self.assertFalse(response.valid)

    def test_request_transfer_exit_wrongservice_exception(self):
        self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

    def _test_request_transfer_exit_wrongservice_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.request_transfer_exit(data=b'\x12\x34\x56')

    def test_request_transfer_exit_wrongservice_no_exception(self):
        self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

    def _test_request_transfer_exit_wrongservice_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False	
        response = self.udsclient.request_transfer_exit(data=b'\x12\x34\x56')
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_bad_param(self):
        pass

    def _test_bad_param(self):
        with self.assertRaises(ValueError):
            self.udsclient.request_transfer_exit(data=123)

        with self.assertRaises(ValueError):
            self.udsclient.request_transfer_exit(data="asd")



from test.UdsTest import UdsTest
from udsoncan.connections import *
from test.stub import StubbedIsoTPSocket
import socket
import threading
import time
import unittest

try:
    _STACK_UNVAILABLE_REASON = ''
    _interface_name = 'vcan0'
    import isotp
    import can
    s = isotp.socket()
    s.bind(_interface_name,rxid=1,txid=2)
    s.close()
    _STACK_POSSIBLE = True
except Exception as e:
    _STACK_UNVAILABLE_REASON = str(e)
    _STACK_POSSIBLE = False

try:
    from aioisotp.sync import SyncISOTPNetwork
    _AISOTP_POSSIBLE = True
except Exception as e:
    _AISOTP_POSSIBLE = False

class TestIsoTPSocketConnection(UdsTest):

    def setUp(self):
        self.tpsock1 = StubbedIsoTPSocket(timeout=0.1)
        self.tpsock2 = StubbedIsoTPSocket(timeout=0.1)

    def test_open(self):
        conn = IsoTPSocketConnection(interface='vcan0', rxid=0x001, txid=0x002, tpsock=self.tpsock1, name='unittest')
        self.assertFalse(conn.is_open())
        conn.open()
        self.assertTrue(conn.is_open())
        conn.close()
        self.assertFalse(conn.is_open())

    def test_transmit(self):
        conn1 = IsoTPSocketConnection(interface='vcan0', rxid=0x100, txid=0x101, tpsock=self.tpsock1, name='unittest')
        conn2 = IsoTPSocketConnection(interface='vcan0', rxid=0x101, txid=0x100, tpsock=self.tpsock2, name='unittest')

        with conn1.open():
            with conn2.open():
                payload1 = b"\x00\x01\x02\x03\x04"
                conn1.send(payload1)
                payload2 = conn2.wait_frame(timeout=0.3)
                self.assertEqual(payload1, payload2)

class TestSocketConnection(UdsTest):
    def server_sock_thread_task(self):
        self.thread_started=True
        self.sock1, addr = self.server_sock.accept()

    def setUp(self):
        self.thread_started = False
        self.server_sock_thread = threading.Thread(target=self.server_sock_thread_task)

        self.server_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_sock.setblocking(False)
        self.sock1 = None
        self.sock2 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self.server_sock.settimeout(0.5)

        self.server_sock.bind(('127.0.0.1', 0))
        self.server_sock.listen(1)
        self.server_sock_thread.start()

        t1 = time.time()
        while not self.thread_started:
            if (time.time() - t1) > 0.5:
                raise RuntimeError('Timeout while connecting sockets together.')
            time.sleep(0.01)
        time.sleep(0.01)

        self.sock2.connect(self.server_sock.getsockname())
        t1 = time.time()
        while self.sock1 is None:
            if (time.time() - t1) > 0.5:
                raise RuntimeError('Timeout while connecting sockets together.')

    def tearDown(self):
        if isinstance(self.sock1, socket.socket):
            self.sock1.close()

        if isinstance(self.sock2, socket.socket):
            self.sock2.close()

        if isinstance(self.server_sock, socket.socket):
            self.server_sock.close()

    def test_open(self):
        conn = SocketConnection(self.sock1, name='unittest')
        self.assertFalse(conn.is_open())
        conn.open()
        self.assertTrue(conn.is_open())
        conn.close()
        self.assertFalse(conn.is_open())

    def test_transmit(self):
        conn1 = SocketConnection(self.sock1, name='unittest')
        conn2 = SocketConnection(self.sock2, name='unittest')

        with conn1.open():
            with conn2.open():
                payload1 = b"\x00\x01\x02\x03\x04"
                conn1.send(payload1)
                payload2 = conn2.wait_frame(timeout=1, exception=True)
                self.assertEqual(payload1, payload2)

class TestQueueConnection(UdsTest):
    def setUp(self):
        self.conn = QueueConnection(name='unittest')
        self.conn.open()

    def tearDown(self):
        self.conn.close()

    def test_open(self):
        self.assertTrue(self.conn.is_open())

    def test_receive(self):
        payload = b"\x00\x01\x02\x03"
        self.conn.fromuserqueue.put(payload)
        frame = self.conn.wait_frame()
        self.assertEqual(frame, payload)

    def test_send(self):
        payload = b"\x00\x01\x02\x03"
        self.conn.send(payload)
        frame = self.conn.touserqueue.get()
        self.assertEqual(frame, payload)

    def test_truncate(self):
        payload = b"\x00\x01\x02\x03"*5000
        self.conn.send(payload)
        frame = self.conn.touserqueue.get()
        self.assertEqual(len(frame), 4095)
        self.assertEqual(frame, payload[0:4095])

        self.conn.fromuserqueue.put(payload)
        frame = self.conn.wait_frame()

        self.assertEqual(len(frame), 4095)
        self.assertEqual(frame, payload[0:4095])

    def test_reopen(self):
        payload = b"\x00\x01\x02\x03"
        self.conn.send(payload)
        self.conn.fromuserqueue.put(payload)

        self.conn.close()
        self.conn.open()

        with self.assertRaises(TimeoutException):
            self.conn.wait_frame(timeout=0.05, exception=True)

        self.assertTrue(self.conn.touserqueue.empty())

@unittest.skipIf(_STACK_POSSIBLE == False, 'Cannot test TestPythonIsoTpConnection. %s' % _STACK_UNVAILABLE_REASON)
class TestPythonIsoTpConnection(UdsTest):
    def __init__(self, *args, **kwargs):
        UdsTest.__init__(self, *args, **kwargs)
        if not hasattr(self.__class__, '_next_id'):
            self.__class__._next_id=1

        self.stack_txid = self.__class__._next_id
        self.stack_rxid = self.__class__._next_id +1
        self.__class__._next_id += 2

    def make_bus(self):
        return can.interface.Bus(bustype='socketcan', channel='vcan0', bitrate=500000, receive_own_messages=True)

    def setUp(self):
        self.vcan0_bus = self.make_bus()
        addr = isotp.Address(isotp.AddressingMode.Normal_11bits, rxid=self.stack_rxid, txid=self.stack_txid)
        self.conn = PythonIsoTpConnection(isotp.CanStack(bus=self.vcan0_bus, address=addr), name='unittest')
        self.conn.open()

    def test_open(self):
        self.assertTrue(self.conn.is_open())

    def test_receive(self):
        self.vcan0_bus.send(can.Message(arbitration_id = self.stack_rxid, data =  b"\x03\x01\x02\x03", is_extended_id = False))
        frame = self.conn.wait_frame(timeout=1)
        self.assertEqual(frame, b"\x01\x02\x03")

    def test_send(self):
        self.conn.send(b"\xAA\xBB\xCC\xDD\xEE\xFF")
        t1 = time.time()
        msg = self.vcan0_bus.recv(1)
        self.assertIsNotNone(msg)
        self.assertEqual(msg.data, b'\x06\xAA\xBB\xCC\xDD\xEE\xFF')

    def test_reopen(self):
        self.conn.send(b"\x0A\x0B\x0C\x0D")
        self.vcan0_bus.send(can.Message(arbitration_id = self.stack_rxid, data =  b"\x03\x01\x02\x03", is_extended_id = False))
        self.conn.close()
        self.vcan0_bus.shutdown()
        self.vcan0_bus = self.make_bus()
        self.conn.open(bus=self.vcan0_bus)

        with self.assertRaises(TimeoutException):
            self.conn.wait_frame(timeout=0.05, exception=True)

        self.assertIsNone(self.vcan0_bus.recv(0))

    def tearDown(self):
        self.conn.close()
        self.vcan0_bus.shutdown()

@unittest.skipIf(_AISOTP_POSSIBLE == False, "aisotp module is not present.")
class TestSyncAioIsotpConnection(UdsTest):

    def test_open(self):
        conn0 = SyncAioIsotpConnection(interface="virtual", channel=0, bitrate=500000, rx_id=0x123, tx_id=0x456, name="unittest")

        conn0.open()
        self.assertTrue(conn0.is_open())
        conn0.close()
        self.assertFalse(conn0.is_open())

        conn0.open()
        self.assertTrue(conn0.is_open())
        conn0.close()
        self.assertFalse(conn0.is_open())

    def test_transmit(self):
        conn0 = SyncAioIsotpConnection(interface="virtual", channel=0, bitrate=500000, rx_id=0x123, tx_id=0x456, name="unittest")
        conn1 = SyncAioIsotpConnection(interface="virtual", channel=0, bitrate=500000, rx_id=0x456, tx_id=0x123, name="unittest")
    
        with conn0.open():
            with conn1.open():
                tx_data = bytes([i for i in range(256)])
                conn0.send(tx_data)

                rx_data = conn1.wait_frame()
                self.assertTrue(tx_data == rx_data)



from udsoncan.client import Client
from udsoncan import services, MemoryLocation, DataFormatIdentifier
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest

class TestRequestDownload(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def test_request_download_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x34\x00\x12\x12\x34\xFF")
        self.conn.fromuserqueue.put(b"\x74\x20\xab\xcd")	# Positive response

    def _test_request_download_success(self):
        memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
        response = self.udsclient.request_download(memory_location=memloc)
        self.assertEqual(response.service_data.max_length,0xabcd)

    def test_request_download_success_spr_no_effect(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x34\x00\x12\x12\x34\xFF")
        self.conn.fromuserqueue.put(b"\x74\x20\xab\xcd")	# Positive response

    def _test_request_download_success_spr_no_effect(self):
        with self.udsclient.suppress_positive_response:
            memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
            response = self.udsclient.request_download(memory_location=memloc)
            self.assertEqual(response.service_data.max_length,0xabcd)

    def test_request_download_config_format(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x34\x00\x24\x00\x00\x12\x34\x00\xFF")	# dfi = 24 and 0 padding
        self.conn.fromuserqueue.put(b"\x74\x20\xab\xcd")	# Positive response

    def _test_request_download_config_format(self):
        self.udsclient.set_configs({'server_address_format':32, 'server_memorysize_format':16})
        memloc = MemoryLocation(address=0x1234, memorysize=0xFF)
        response = self.udsclient.request_download(memory_location=memloc)
        self.assertEqual(response.service_data.max_length,0xabcd)

    def test_request_download_success_lfid(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x34\x00\x12\x12\x34\xFF")
        self.conn.fromuserqueue.put(b"\x74\x23\xab\xcd")	# Positive response

    def _test_request_download_success_lfid(self):
        memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
        response = self.udsclient.request_download(memory_location=memloc)
        self.assertEqual(response.service_data.max_length,0xabcd)

    def test_request_download_success_dfi(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x34\x52\x12\x12\x34\xFF")
        self.conn.fromuserqueue.put(b"\x74\x20\xab\xcd")	# Positive response

    def _test_request_download_success_dfi(self):
        memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
        dfi  =DataFormatIdentifier(compression=5, encryption=2)
        response = self.udsclient.request_download(memory_location=memloc, dfi=dfi)
        self.assertEqual(response.service_data.max_length,0xabcd)

    def test_incomplete_nblock_response_exception(self):
        self.wait_request_and_respond(b"\x74\x40\xab\xcd")	# Missing 2 bytes to complete number of block

    def _test_incomplete_nblock_response_exception(self):
        memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
        with self.assertRaises(InvalidResponseException):
            self.udsclient.request_download(memory_location=memloc)

    def test_incomplete_nblock_response_no_exception(self):
        self.wait_request_and_respond(b"\x74\x40\xab\xcd")	# Missing 2 bytes to complete number of block

    def _test_incomplete_nblock_response_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
        response = self.udsclient.request_download(memory_location=memloc)
        self.assertFalse(response.valid)

    def test_request_download_invalidservice_exception(self):
        self.wait_request_and_respond(b"\x00\x20\x12\x34") #Inexistent Service

    def _test_request_download_invalidservice_exception(self):
        memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.request_download(memory_location=memloc)

    def test_request_download_invalidservice_no_exception(self):
        self.wait_request_and_respond(b"\x00\x20\x12\x34") #Inexistent Service

    def _test_request_download_invalidservice_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False		
        memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
        response = self.udsclient.request_download(memory_location=memloc)
        self.assertFalse(response.valid)

    def test_request_download_wrongservice_exception(self):
        self.wait_request_and_respond(b"\x7E\x20\x12\x34") # Valid but wrong service (Tester Present)

    def _test_request_download_wrongservice_exception(self):
        memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.request_download(memory_location=memloc)

    def test_request_download_wrongservice_no_exception(self):
        self.wait_request_and_respond(b"\x7E\x20\x12\x34") # Valid but wrong service (Tester Present)

    def _test_request_download_wrongservice_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False	
        memloc = MemoryLocation(address=0x1234, memorysize=0xFF, address_format=16, memorysize_format=8)
        response = self.udsclient.request_download(memory_location=memloc)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_bad_params(self):
        pass

    def _test_bad_params(self):
        with self.assertRaises(ValueError) as handle:
            self.udsclient.request_download(memory_location=1)

        with self.assertRaises(ValueError) as handle:
            self.udsclient.request_download(memory_location="asd")



import inspect

from udsoncan.ResponseCode import ResponseCode
from abc import ABC

from typing import Type, List, Optional


class BaseSubfunction:

    @classmethod
    def get_name(cls, subfn_id: int) -> str:
        attributes = inspect.getmembers(cls, lambda a: not (inspect.isroutine(a)))
        subfn_list = [a for a in attributes if not (a[0].startswith('__') and a[0].endswith('__'))]

        for subfn in subfn_list:
            if isinstance(subfn[1], int):
                if subfn[1] == subfn_id:  # [1] is value
                    return subfn[0] 		# [0] is property name
            elif isinstance(subfn[1], tuple):
                if subfn_id >= subfn[1][0] or subfn_id <= subfn[1][1]:
                    return subfn[0]
        name = cls.__name__ if not hasattr(cls, '__pretty_name__') else cls.__pretty_name__
        return 'Custom %s' % name


class BaseService(ABC):

    always_valid_negative_response = [
        ResponseCode.GeneralReject,
        ResponseCode.ServiceNotSupported,
        ResponseCode.ResponseTooLong,
        ResponseCode.BusyRepeatRequest,
        ResponseCode.NoResponseFromSubnetComponent,
        ResponseCode.FailurePreventsExecutionOfRequestedAction,
        ResponseCode.SecurityAccessDenied,  # ISO-14229:2006 Table A.1:  "Besides the mandatory use of this negative response code as specified in the applicable services within ISO 14229, this negative response code can also be used for any case where security is required and is not yet granted to perform the required service."
        ResponseCode.RequestCorrectlyReceived_ResponsePending,
        ResponseCode.ServiceNotSupportedInActiveSession,
        ResponseCode.ResourceTemporarilyNotAvailable
    ]

    _sid: int
    _use_subfunction: bool
    supported_negative_response: List[int]

    @classmethod  # Returns the service ID used for a client request
    def request_id(cls) -> int:
        return cls._sid

    @classmethod  # Returns the service ID used for a server response
    def response_id(cls) -> int:
        return cls._sid + 0x40

    @staticmethod
    def __get_all_subclasses(cls) -> List[Type["BaseService"]]:
        # this gets all subclasses and returns a list where the "most original" subclasses are listed in front of the other subclasses
        # This enables subclasses of any BaseService outside of udsoncan.services to be available, enabling specialization of calls in
        # cases where a CAN message is similar to one found in official UDS documentation but has a different service ID
        # This also allows for custom UDS service creation where a nonstandard extension is more easily played ontop of the protocol
        lst = []
        lst.extend(cls.__subclasses__())
        for x in cls.__subclasses__():
            subclasses = BaseService.__get_all_subclasses(x)
            if len(subclasses) > 0:
                lst.extend(subclasses)
        return lst

    @classmethod  # Returns an instance of the service identified by the service ID (Request)
    def from_request_id(cls, given_id: int) -> Optional[Type["BaseService"]]:
        classes = BaseService.__get_all_subclasses(cls)
        for obj in classes:
            if obj.request_id() == given_id:
                return obj
        return None

    @classmethod  # Returns an instance of the service identified by the service ID (Response)
    def from_response_id(cls, given_id: int) -> Optional[Type["BaseService"]]:
        classes = BaseService.__get_all_subclasses(cls)
        for obj in classes:
            if obj.response_id() == int(given_id):
                return obj
        return None

    # Default subfunction ID for service that does not implement subfunction_id().
    def subfunction_id(self) -> int:
        return 0

    @classmethod  # Tells if this service includes a subfunction byte
    def use_subfunction(cls) -> bool:
        if hasattr(cls, '_use_subfunction'):
            return cls._use_subfunction
        else:
            return True

    @classmethod
    def has_response_data(cls) -> bool:
        if hasattr(cls, '_no_response_data'):
            return False if cls._no_response_data else True
        else:
            return True

    @classmethod  # Returns the service name. Shortcut that works on class and instances
    def get_name(cls) -> str:
        return cls.__name__

    @classmethod  # Tells if the given response code is expected for this service according to UDS standard.
    def is_supported_negative_response(cls, code: int) -> bool:
        supported = False
        if code in cls.supported_negative_response:
            supported = True

        if code in cls.always_valid_negative_response:
            supported = True

        # As specified by Annex A, negative response code ranging above 0x7F can be used anytime if the service can return ConditionNotCorrect
        if code >= 0x80 and code < 0xFF and ResponseCode.ConditionsNotCorrect in cls.supported_negative_response:
            supported = True

        # ISO-14229:2006 Table A.1 : "This response code shall be supported by each diagnostic service with a subfunction parameter"
        if code == ResponseCode.SubFunctionNotSupportedInActiveSession and cls.use_subfunction():
            supported = True

        return supported


class BaseResponseData:
    def __init__(self, service_class):
        if not issubclass(service_class, BaseService):
            raise ValueError('service_class must be a service class')

        self.service_class = service_class

    def __repr__(self):
        return '<%s (%s) at 0x%08x>' % (self.__class__.__name__, self.service_class.__name__, id(self))



__all__ = [
    'DataIdentifier',
    'CodecDefinition',
    'check_did_config',
    'make_did_codec_from_config'
]

import inspect
from udsoncan.exceptions import ConfigError
from udsoncan.typing import CodecDefinition, DIDConfig, IOConfigEntry
from udsoncan.common.DidCodec import DidCodec

from typing import Dict, List, Union, Optional, cast


class DataIdentifier:
    """
    Defines a list of constants that are data identifiers defined by the UDS standard.
    This class provides no functionality apart from defining these constants
    """
    BootSoftwareIdentification = 0xF180
    ApplicationSoftwareIdentification = 0xF181
    ApplicationDataIdentification = 0xF182
    BootSoftwareFingerprint = 0xF183
    ApplicationSoftwareFingerprint = 0xF184
    ApplicationDataFingerprint = 0xF185
    ActiveDiagnosticSession = 0xF186
    VehicleManufacturerSparePartNumber = 0xF187
    VehicleManufacturerECUSoftwareNumber = 0xF188
    VehicleManufacturerECUSoftwareNumber = 0xF188
    VehicleManufacturerECUSoftwareVersionNumber = 0xF189
    SystemSupplierIdentifier = 0xF18A
    ECUManufacturingDate = 0xF18B
    ECUSerialNumber = 0xF18C
    SupportedFunctionalUnits = 0xF18D
    VehicleManufacturerKitAssemblyPartNumber = 0xF18E
    ISOSAEReservedStandardized = 0xF18F
    VIN = 0xF190
    VehicleManufacturerECUHardwareNumber = 0xF191
    SystemSupplierECUHardwareNumber = 0xF192
    SystemSupplierECUHardwareVersionNumber = 0xF193
    SystemSupplierECUSoftwareNumber = 0xF194
    SystemSupplierECUSoftwareVersionNumber = 0xF195
    ExhaustRegulationOrTypeApprovalNumber = 0xF196
    SystemNameOrEngineType = 0xF197
    RepairShopCodeOrTesterSerialNumber = 0xF198
    ProgrammingDate = 0xF199
    CalibrationRepairShopCodeOrCalibrationEquipmentSerialNumber = 0xF19A
    CalibrationDate = 0xF19B
    CalibrationEquipmentSoftwareNumber = 0xF19C
    ECUInstallationDate = 0xF19D
    ODXFile = 0xF19E
    Entity = 0xF19F

    @classmethod
    def name_from_id(cls, did: int) -> Optional[str]:
        # As defined by ISO-14229:2006, Annex F
        if not isinstance(did, int) or did < 0 or did > 0xFFFF:
            raise ValueError('Data IDentifier must be a valid integer between 0 and 0xFFFF')

        if did >= 0x0000 and did <= 0x00FF:
            return 'ISOSAEReserved'
        if did >= 0x0100 and did <= 0xEFFF:
            return 'VehicleManufacturerSpecific'
        if did >= 0xF000 and did <= 0xF00F:
            return 'NetworkConfigurationDataForTractorTrailerApplicationDataIdentifier'
        if did >= 0xF010 and did <= 0xF0FF:
            return 'VehicleManufacturerSpecific'
        if did >= 0xF100 and did <= 0xF17F:
            return 'IdentificationOptionVehicleManufacturerSpecificDataIdentifier'

        if did == 0xF180:
            return 'BootSoftwareIdentificationDataIdentifier'
        if did == 0xF181:
            return 'ApplicationSoftwareIdentificationDataIdentifier'
        if did == 0xF182:
            return 'ApplicationDataIdentificationDataIdentifier'
        if did == 0xF183:
            return 'BootSoftwareFingerprintDataIdentifier'
        if did == 0xF184:
            return 'ApplicationSoftwareFingerprintDataIdentifier'
        if did == 0xF185:
            return 'ApplicationDataFingerprintDataIdentifier'
        if did == 0xF186:
            return 'ActiveDiagnosticSessionDataIdentifier'
        if did == 0xF187:
            return 'VehicleManufacturerSparePartNumberDataIdentifier'
        if did == 0xF188:
            return 'VehicleManufacturerECUSoftwareNumberDataIdentifier'
        if did == 0xF188:
            return 'VehicleManufacturerECUSoftwareNumberDataIdentifier'
        if did == 0xF189:
            return 'VehicleManufacturerECUSoftwareVersionNumberDataIdentifier'
        if did == 0xF18A:
            return 'SystemSupplierIdentifierDataIdentifier'
        if did == 0xF18B:
            return 'ECUManufacturingDateDataIdentifier'
        if did == 0xF18C:
            return 'ECUSerialNumberDataIdentifier'
        if did == 0xF18D:
            return 'SupportedFunctionalUnitsDataIdentifier'
        if did == 0xF18E:
            return 'VehicleManufacturerKitAssemblyPartNumberDataIdentifier'
        if did == 0xF18F:
            return 'ISOSAEReservedStandardized'
        if did == 0xF190:
            return 'VINDataIdentifier'
        if did == 0xF191:
            return 'VehicleManufacturerECUHardwareNumberDataIdentifier'
        if did == 0xF192:
            return 'SystemSupplierECUHardwareNumberDataIdentifier'
        if did == 0xF193:
            return 'SystemSupplierECUHardwareVersionNumberDataIdentifier'
        if did == 0xF194:
            return 'SystemSupplierECUSoftwareNumberDataIdentifier'
        if did == 0xF195:
            return 'SystemSupplierECUSoftwareVersionNumberDataIdentifier'
        if did == 0xF196:
            return 'ExhaustRegulationOrTypeApprovalNumberDataIdentifier'
        if did == 0xF197:
            return 'SystemNameOrEngineTypeDataIdentifier'
        if did == 0xF198:
            return 'RepairShopCodeOrTesterSerialNumberDataIdentifier'
        if did == 0xF199:
            return 'ProgrammingDateDataIdentifier'
        if did == 0xF19A:
            return 'CalibrationRepairShopCodeOrCalibrationEquipmentSerialNumberDataIdentifier'
        if did == 0xF19B:
            return 'CalibrationDateDataIdentifier'
        if did == 0xF19C:
            return 'CalibrationEquipmentSoftwareNumberDataIdentifier'
        if did == 0xF19D:
            return 'ECUInstallationDateDataIdentifier'
        if did == 0xF19E:
            return 'ODXFileDataIdentifier'
        if did == 0xF19F:
            return 'EntityDataIdentifier'

        if did >= 0xF1A0 and did <= 0xF1EF:
            return 'IdentificationOptionVehicleManufacturerSpecific'
        if did >= 0xF1F0 and did <= 0xF1FF:
            return 'IdentificationOptionSystemSupplierSpecific'
        if did >= 0xF200 and did <= 0xF2FF:
            return 'PeriodicDataIdentifier'
        if did >= 0xF300 and did <= 0xF3FF:
            return 'DynamicallyDefinedDataIdentifier'
        if did >= 0xF400 and did <= 0xF4FF:
            return 'OBDDataIdentifier'
        if did >= 0xF500 and did <= 0xF5FF:
            return 'OBDDataIdentifier'
        if did >= 0xF600 and did <= 0xF6FF:
            return 'OBDMonitorDataIdentifier'
        if did >= 0xF700 and did <= 0xF7FF:
            return 'OBDMonitorDataIdentifier'
        if did >= 0xF800 and did <= 0xF8FF:
            return 'OBDInfoTypeDataIdentifier'
        if did >= 0xF900 and did <= 0xF9FF:
            return 'TachographDataIdentifier'
        if did >= 0xFA00 and did <= 0xFA0F:
            return 'AirbagDeploymentDataIdentifier'
        if did >= 0xFA10 and did <= 0xFAFF:
            return 'SafetySystemDataIdentifier'
        if did >= 0xFB00 and did <= 0xFCFF:
            return 'ReservedForLegislativeUse'
        if did >= 0xFD00 and did <= 0xFEFF:
            return 'SystemSupplierSpecific'
        if did >= 0xFF00 and did <= 0xFFFF:
            return 'ISOSAEReserved'

        return None


def check_did_config(didlist: Union[int, List[int]], didconfig: Optional[Dict]) -> DIDConfig:
    """Make sure that the actual client configuration contains valid definitions for given Data Identifiers"""
    if didconfig is None:
        raise ConfigError("didconfig is not set")
    didlist = [didlist] if not isinstance(didlist, list) else didlist
    if 'data_identifiers' in didconfig:
        didconfig = didconfig['data_identifiers']

    assert didconfig is not None
    for did in didlist:
        if did not in didconfig:
            raise ConfigError(did, msg='Actual data identifier configuration contains no definition for data identifier 0x%04x' % did)

    return cast(DIDConfig, didconfig)


def make_did_codec_from_config(didconfig: Union[CodecDefinition, IOConfigEntry]) -> DidCodec:
    if isinstance(didconfig, DidCodec):  # the given object already is a DidCodec instance
        return didconfig

    # The definition of the codec is a class. Returns an instance of this codec.
    if inspect.isclass(didconfig) and issubclass(didconfig, DidCodec):
        return didconfig()

    # It could be that the codec is in a dict. (for io_control)
    if isinstance(didconfig, dict) and 'codec' in didconfig:
        return make_did_codec_from_config(didconfig['codec'])

    # The codec can be defined by a struct pack/unpack string
    if isinstance(didconfig, str):
        if len(didconfig) == 0:
            raise ValueError("pack/unpack string given for Codec config should not be empty.")
        return DidCodec(packstr=didconfig)

    raise ValueError('Given codec of type %s is not a valid DidCodec' % (type(didconfig)))



from udsoncan.client import Client
from udsoncan import services
from udsoncan.exceptions import *

from test.ClientServerTest import ClientServerTest

class TestTransferData(ClientServerTest):
    def __init__(self, *args, **kwargs):
        ClientServerTest.__init__(self, *args, **kwargs)

    def test_transfer_data_success(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x36\x22\x12\x34\x56")
        self.conn.fromuserqueue.put(b"\x76\x22\x89\xab\xcd\xef")	# Positive response

    def _test_transfer_data_success(self):
        response = self.udsclient.transfer_data(sequence_number=0x22, data=b'\x12\x34\x56')
        self.assertEqual(response.service_data.sequence_number_echo, 0x22)
        self.assertEqual(response.service_data.parameter_records, b'\x89\xab\xcd\xef')

    def test_transfer_data_success_spr_no_effect(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x36\x22\x12\x34\x56")
        self.conn.fromuserqueue.put(b"\x76\x22\x89\xab\xcd\xef")	# Positive response

    def _test_transfer_data_success_spr_no_effect(self):
        with self.udsclient.suppress_positive_response:
            response = self.udsclient.transfer_data(sequence_number=0x22, data=b'\x12\x34\x56')
            self.assertEqual(response.service_data.sequence_number_echo, 0x22)
            self.assertEqual(response.service_data.parameter_records, b'\x89\xab\xcd\xef')

    def test_transfer_data_no_data_ok(self):
        request = self.conn.touserqueue.get(timeout=0.2)
        self.assertEqual(request, b"\x36\x22")
        self.conn.fromuserqueue.put(b"\x76\x22")	# Positive response

    def _test_transfer_data_no_data_ok(self):
        response = self.udsclient.transfer_data(sequence_number=0x22)
        self.assertEqual(response.service_data.sequence_number_echo, 0x22)
        self.assertEqual(response.service_data.parameter_records, b'')	

    def test_transfer_data_denied_exception(self):
        self.wait_request_and_respond(b"\x7F\x36\x73") # wrong block sequence number

    def _test_transfer_data_denied_exception(self):
        with self.assertRaises(NegativeResponseException) as handle:
            self.udsclient.transfer_data(sequence_number=0x55)
        response = handle.exception.response

        self.assertTrue(response.valid)
        self.assertFalse(response.positive)
        self.assertTrue(issubclass(response.service, services.TransferData))
        self.assertEqual(response.code, 0x73)

    def test_transfer_data_denied_no_exception(self):
        self.wait_request_and_respond(b"\x7F\x36\x73") # wrong block sequence number

    def _test_transfer_data_denied_no_exception(self):
        self.udsclient.config['exception_on_negative_response'] = False
        response = self.udsclient.transfer_data(sequence_number=0x55)

        self.assertTrue(response.valid)
        self.assertFalse(response.positive)
        self.assertTrue(issubclass(response.service, services.TransferData))
        self.assertEqual(response.code, 0x73)

    def test_transfer_data_bad_sequence_number_exception(self):
        self.wait_request_and_respond(b"\x76\x23\x89\xab\xcd\xef")	# Positive response

    def _test_transfer_data_bad_sequence_number_exception(self):
        with self.assertRaises(UnexpectedResponseException):
            self.udsclient.transfer_data(sequence_number=0x22, data=b'\x12\x34\x56')

    def test_transfer_data_bad_sequence_number_no_exception(self):
        self.wait_request_and_respond(b"\x76\x23\x89\xab\xcd\xef")	# Positive response

    def _test_transfer_data_bad_sequence_number_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.transfer_data(sequence_number=0x22, data=b'\x12\x34\x56')
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_transfer_data_invalidservice_exception(self):
        self.wait_request_and_respond(b"\x00\x55") #Inexistent Service

    def _test_transfer_data_invalidservice_exception(self):
        with self.assertRaises(InvalidResponseException) as handle:
            self.udsclient.transfer_data(sequence_number=0x55)

    def test_transfer_data_invalidservice_no_exception(self):
        self.wait_request_and_respond(b"\x00\x55") #Inexistent Service

    def _test_transfer_data_invalidservice_no_exception(self):
        self.udsclient.config['exception_on_invalid_response'] = False
        response = self.udsclient.transfer_data(sequence_number=0x55)
        self.assertFalse(response.valid)

    def test_transfer_data_wrongservice_exception(self):
        self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

    def _test_transfer_data_wrongservice_exception(self):
        with self.assertRaises(UnexpectedResponseException) as handle:
            self.udsclient.transfer_data(sequence_number=0x55)

    def test_transfer_data_wrongservice_no_exception(self):
        self.wait_request_and_respond(b"\x7E\x00") # Valid but wrong service (Tester Present)

    def _test_transfer_data_wrongservice_no_exception(self):
        self.udsclient.config['exception_on_unexpected_response'] = False
        response = self.udsclient.transfer_data(sequence_number=0x55)
        self.assertTrue(response.valid)
        self.assertTrue(response.unexpected)

    def test_bad_param(self):
        pass

    def _test_bad_param(self):
        with self.assertRaises(ValueError):
            self.udsclient.transfer_data(sequence_number=-1)

        with self.assertRaises(ValueError):
            self.udsclient.transfer_data(sequence_number=0x100)

        with self.assertRaises(ValueError):
            self.udsclient.transfer_data(sequence_number=1, data=123)

        with self.assertRaises(ValueError):
            self.udsclient.transfer_data(sequence_number=1, data='asd')



import struct
from udsoncan import Request, Response, latest_standard
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
import udsoncan.tools as tools
from typing import Optional, cast


class ClearDiagnosticInformation(BaseService):
    _sid = 0x14
    _use_subfunction = False
    _no_response_data = True

    supported_negative_response = [ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestOutOfRange
                                   ]

    class ResponseData(BaseResponseData):
        """
        Empty object
        """

        def __init__(self):
            super().__init__(ClearDiagnosticInformation)

    class InterpretedResponse(Response):
        service_data: "ClearDiagnosticInformation.ResponseData"

    @classmethod
    def make_request(cls, group: int = 0xFFFFFF, memory_selection: Optional[int] = None, standard_version: int = latest_standard) -> Request:
        """
        Generates a request for ClearDiagnosticInformation

        :param group: DTC mask ranging from 0 to 0xFFFFFF. 0xFFFFFF means all DTCs
        :type group: int

        :raises ValueError: If parameters are out of range, missing or wrong type
        """
        tools.validate_int(group, min=0, max=0xFFFFFF, name='Group of DTC')
        request = Request(service=cls)
        hb = (group >> 16) & 0xFF
        mb = (group >> 8) & 0xFF
        lb = (group >> 0) & 0xFF
        request_data = struct.pack("BBB", hb, mb, lb)

        # Introduced in ISO-14229-1:2020
        if memory_selection is not None:
            if standard_version < 2020:
                raise NotImplementedError(
                    'ClearDiagnosticInformation with Memory Selection is only possible with 2020 version of the standard or above.')

            tools.validate_int(memory_selection, min=0, max=0xFF, name='Memory Selection')
            request_data += struct.pack("B", memory_selection)
        request.data = request_data
        return request

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`ClearDiagnosticInformation.ResponseData<udsoncan.services.ClearDiagnosticInformation.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`
        """
        response.service_data = cls.ResponseData()
        return cast(ClearDiagnosticInformation.InterpretedResponse, response)
